from typing import Literal

PLAYER_ACTION_ACTION = Literal['SetHotbar', 'SetReducedDebug', 'CloseInv', 'GiveItems', 'NoKeepInv', 'SetHandCrafting', 'BossBar', 'SetBossBar', 'ParticleSphere', 'SetVelocity', 'Particle', 'AddInvRow', 'NoNatRegen', 'DisplayLightning', 'Damage', 'SendAnimation', 'SetXPProg', 'SetInventory', 'TpSequence', 'Heal', 'SetSpawnPoint', 'SetInventoryKept', 'LaunchUp', 'GetTargetEntity', 'ForceFlight', 'LoadInv', 'ChatStyle', 'ChatColor', 'Kick', 'ProjColl', 'MiscAttribute', 'SpectateTarget', 'HurtAnimation', 'SurvivalMode', 'GmSurvival', 'DisplayBellRing', 'SetStatus', 'SetCursorItem', 'SetAbsorption', 'SetFireTicks', 'CombatAttribute', 'SetGamemode', 'RemoveInvRow', 'WakeUpAnimation', 'DisableBlocks', 'SetScoreObj', 'L SetHealth', 'ParticleEffect', 'ClearInv', 'SetFreezeTicks', 'SetGliding', 'SetRotation', 'ClearItems', 'OpenSign', 'SetFlying', 'DisplayBlockOpen', 'SetHandItem', 'SendAdvancement', 'ClearChat', 'SetMenuItem', 'LaunchToward', 'SetArmor', 'DisplayGateway', 'GiveSaturation', 'DisplayEquipment', 'GiveExp', 'FaceLocation', 'ClearScoreboard', 'ActionBar', 'SetChatTag', 'ShiftWorldBorder', 'DisplaySignText', 'SetSpeed', 'ExpandInv', 'LaunchProj', 'NoProjColl', 'ShowDisguise', 'ParticleCuboidA', 'PlaySound', 'SetCompass', 'MobDisguise', 'EnableBlocks', 'OpenBlockInv', 'ParticleCircleA', 'RemoveBossBar', 'ClearBars', 'SetEquipment', 'GiveRngItem', 'SetDropsEnabled', 'SendToPlot', 'RemovePotion', 'RemoveEffect', 'DisplayFracture', 'SetEntityHidden', 'SetSidebar', 'AllowDrops', 'Vibration', 'SetSlot', 'ParticleRay', 'ParticleCuboid', 'SendMessageSeq', 'SendDialogue', 'SetNamePrefix', 'ClearDispBlock', 'SetRainLevel', 'Undisguise', 'ParticleSpiralA', 'InstantRespawn', 'SetScore', 'SetNameColor', 'ReachAttribute', 'SetAtkSpeed', 'DisablePvp', 'SetTickRate', 'PlayEntitySound', 'ReplaceProj', 'SetExp', 'MiningAttribute', 'KBAttribute', 'MovementAttribute', 'ParticleSpiral', 'FallingAttribute', 'SetAllowFlight', 'SetMaxHealth', ' RemoveBossBar ', 'SetFogDistance', 'AdventureMode', 'GmAdventure', 'SpectatorMode', 'DispHeadTexture', 'ClearPotions', 'ClearEffects', 'SetTabListInfo', 'EnablePvp', 'PromptPurchase', 'HideDisguise', 'ScoreLineFormat', ' SetBossBar ', 'SetSkin', 'SpectatorCollision', 'SetNameVisible', 'SetInvulTicks', 'EnableFlight', 'SetStingsStuck', 'RemoveScore', 'DisallowDrops', 'SetExhaustion', 'ParticleCircle', 'DisplayBlock', 'RideEntity', 'WeatherRain', 'RmWorldBorder', 'ResourcePack', ' SetInvName ', 'GiveExhaustion', 'Teleport', 'SetAllowPVP', 'DisableFlight', 'SetVisualFire', 'SetDisguiseVisible', 'SetArrowsStuck', 'GetItemCooldown', 'SetItems', 'KeepInv', 'ReplaceItems', 'ReplaceItem', 'SendMessage', 'SetSlotItem', 'PlaySoundSeq', 'ParticleLineA', 'Respawn', 'SetInvName', 'SetItemCooldown', 'SetPlayerWeather', 'SendHover', 'SetShoulder', 'SetAirTicks', 'DisplayPickup', 'SetWorldBorder', 'SetPlayerTime', 'GiveFood', 'NatRegen', 'GivePotion', 'GiveEffect', 'RemoveItems', 'BoostElytra', 'SaveInv', 'OpenBook', 'SetHealth', 'BlockDisguise', 'RollbackBlocks', 'NoDeathDrops', 'WalkSpeed', 'SetCollidable', 'LaunchFwd', 'SetFallDistance', 'CreativeMode', 'AttackAnimation', 'DisplayHologram', ' SetAbsorption ', 'DeathDrops', 'ShowInv', 'SetFoodLevel', 'PlayerDisguise', 'SetSaturation', 'WeatherClear', 'SendTitle', 'Title', 'ScoreDefFormat', 'StopSound', 'HealthAttribute', 'ParticleLine']
ENTITY_ACTION_ACTION = Literal['DispRotationEuler', 'Shear', 'SetVelocity', 'SetGlowSquidDark', 'SetFrogType', 'DispRotAxisAngle', 'Damage', 'SetMobSitting', 'SendAnimation', 'DisableGlowing', 'SetWardenAnger', 'SetHorsePattern', 'Heal', 'SetPandaSadTicks', 'SetItemOwner', 'SetDyeColor', 'LaunchUp', 'SetAge', ' SetName ', 'NoGravity', 'SetArmsRaised', 'SetMoveSpeed', 'SetInvulnerable', 'SetFriction', 'ProjColl', 'ArmorStandTags', 'SetPickupDelay', 'DropItems', 'MiscAttribute', 'SetCreeperPower', 'SetMarker', 'RemoveCustomTag', 'SetAbsorption', 'CreeperCharged', 'SetFireTicks', 'CombatAttribute', 'SetName', 'Jump', 'BDisplayBlock', 'SetFreezeTicks', 'TDisplaySeeThru', 'SetGliding', 'SetRotation', 'SetPandaRolling', 'SetFishPattern', 'SetWolfType', 'DispInterpolation', 'SetHandItem', 'SetEndermanBlock', 'LaunchToward', 'SetArmor', ' SetArmor ', 'GetCustomTag', 'InteractionSize', 'FaceLocation', 'SetCatType', 'SetArrowDamage', 'DisplayScale', 'TDisplayAlign', 'LaunchProj', 'EnableAI', 'DisplayBillboard', 'NoProjColl', 'Tame', 'SetGoatScreaming', 'SetBeeStinger', 'MobDisguise', 'SetMinecartBlock', 'FoxSleeping', 'SetEquipment', 'SetSilenced', 'SetBeeNectar', 'AttachLead', 'RemovePotion', 'RemoveEffect', 'ShearSheep', 'ArmorStandSlots', 'SetAllayDancing', 'SetRabbitType', 'SetSize', 'ShowName', 'SetAngry', 'SetAngerTicks', 'Undisguise', 'SetDeathDrops', 'SetPersistent', 'ProjectileItem', 'SetNameColor', 'SetCarryingChest', 'SetParrotColor', 'DispTranslation', 'Remove', 'TDispBackground', 'DisplayCullingSize', 'HideName', 'SetAxolotlColor', 'GetAllEntityTags', 'SetAI', 'KBAttribute', 'MovementAttribute', 'SetRiptiding', 'SetArrowNoClip', 'FallingAttribute', 'SetProjSource', 'SetFoxLeaping', 'SetPandaGene', 'SetMaxHealth', 'SetFishingTime', 'EndCrystalBeam', 'FrogEat', 'DisplayBrightness', 'SetProfession', 'ClearPotions', 'ClearEffects', 'ArmorStandParts', ' SetNameVisible ', 'SetTarget', 'TDisplayShadow', 'SetNameVisible', 'SetInvulTicks', 'SetShulkerPeek', ' SetPose ', 'SetRearing', 'SetCloudRadius', 'SetGravity', 'DispTPDuration', 'SetWitherInvul', 'Silence', 'SetArrowPierce', 'DisplayShadow', 'InteractResponse', 'UseItem', 'RideEntity', 'DisplayMatrix', 'NoDrops', 'SnifferState', 'EnableGlowing', 'Teleport', 'DisplayGlowColor', 'SetVisualFire', 'SetAge/Size', 'L SetArmor', 'SetSaddle', 'SetBulletTarget', 'TDisplayLineWidth', 'SetDragonPhase', 'SetLlamaColor', 'SetVillagerBiome', 'SetCreeperFuse', 'SetBaby', 'MooshroomType', 'SetInvisible', 'SheepEat', 'SetCatResting', 'GivePotion', 'GiveEffect', 'SetGoatHorns', 'SetGlowing', 'SetPandaOnBack', 'IDisplayModelType', 'SetHealth', 'BlockDisguise', 'SetCollidable', 'ArmorStandPose', 'SetPose', 'LaunchFwd', 'SetFallDistance', 'MoveToLoc', 'TDisplayOpacity', 'IDisplayItem', 'AttackAnimation', 'SnowmanPumpkin', 'SetCustomTag', 'Gravity', 'DisplayViewRange', 'NoAI', 'PlayerDisguise', 'SetItem', 'Explode', 'ExplodeCreeper', 'SetDigging', 'MoveTo', 'SetArrowHitSound', 'SetVexCharging', 'SetVillagerExp', 'IgniteCreeper', 'SetCelebrating', 'TDisplayText', 'SetHorseJump', 'Unsilence', 'HealthAttribute', 'Ram', 'SetFoxType']
GAME_ACTION_ACTION = Literal['FillContainer', 'BreakBlock', 'L PFX Spiral', 'ParticleSphere', 'PFX Sphere', 'ChangeSign', 'WebRequest', 'ClearScBoard', 'HideSidebar', 'SpawnItemDisplay', 'WriteTransaction', 'ParticleSpiral', 'PFX Spiral', 'SetBlockData', 'Firework', 'SetEventDamage', 'SpawnItem', 'SignColor', 'ShulkerBullet', 'FireworkEffect', 'SetContainer', 'SpawnInteraction', 'SetItemInSlot', 'CloneRegion', 'CopyBlocks', 'UncancelEvent', 'SetLecternBook', 'SpawnArmorStand', 'SpawnBlockDisp', 'ClearContainer', 'CancelEvent', 'ParticleEffect', 'Particle FX', 'SpawnFangs', 'SetEventSound', 'SetEventXP', 'LockContainer', 'RemoveScore', 'CreateHologram', 'SetExhaustion', 'ParticleCircle', 'PFX Circle', 'PFX Line [A]', 'ClearItems', 'StartLoop', 'SetFurnaceSpeed', 'BlockDropsOn', 'BoneMeal', 'DebugStackTrace', 'FallingBlock', 'DiscordWebhook', 'TickBlock', 'ReplaceItems', 'SetEventProj', 'Explosion', 'SpawnMob', 'SetBrushableItem', 'ParticleLineA', 'SpawnEnderEye', 'ShowSidebar', 'SpawnPotionCloud', 'LaunchProj', ' SetBlock ', 'SpawnItemDisp', 'SetBlockGrowth', 'Wait', 'SetContainerName', 'SetHead', 'RemoveHologram', 'RemoveItems', 'SpawnRngItem', 'SetRegion', 'SetBlock', 'ParticleCircleA', 'PFX Circle [A]', 'SpawnTNT', 'SpawnExpOrb', 'SetBiome', 'SetEventHeal', 'PFX Path', 'ApplyTransaction', 'ParticleRay', 'PFX Ray', 'GenerateTree', 'StopLoop', 'SetScObj', 'SpawnCrystal', 'SetCampfireItem', 'SpawnTextDisplay', 'SpawnVehicle', 'Lightning', 'ParticleSpiralA', 'PFX Spiral [A]', 'SetScore', 'ParticleCluster', 'PFX Cluster', 'L PFX Cluster', 'BlockDropsOff', 'ParticleLine', 'PFX Line']
SET_VAR_ACTION = Literal['GetItemFood', 'String', 'Text', 'SetParticleType', 'SetItemEnchants', 'ClearItemTag', 'PurgeVars', 'ShiftAllAxes', 'ShiftLocCoord', 'GetParticleMat', 'SetParticleSprd', 'AbsoluteValue', 'RemoveItemAttrs', 'AppendValue', '%', 'ShiftOnVector', 'AddVectorLoc', 'GetItemAttribute', 'ClearDict', '+', 'ShiftRotation', 'RotateLocation', '-', ' GetItemName ', 'GetItemRarity', 'MultiplyVector', '/', 'GetSignText', 'Bitwise', 'GetLecternPage', 'ParseX', 'ShiftOnAxis', 'ShiftAxis', 'ParseY', 'VectorBetween', 'ParseZ', 'GetVectorComp', '=', 'RmText', 'AddItemAttribute', 'GetCenterLoc', 'FindCenter', 'Base64Decode', 'AlignLoc', 'GetSoundVolume', 'RandomNumber', 'GetCrossbowProj', 'GzipDecompress', 'ContainerName', 'Raycast', 'RotateAroundVec', 'SetItemFood', 'GetParticleMotion', 'SetParticleMotion', 'Average', 'WrapNumber', 'ClampLoc', 'SetY', 'SetMapTexture', 'GetBlockData', 'SetX', 'SortDict', 'GetLecternBook', 'GetCustomSound', 'CrossProduct', 'x', 'GetParticleRoll', 'ParseYaw', 'DotProduct', 'SetZ', 'SetArmorTrim', 'PopListValue', 'SetParticleOpac', 'Noise', 'MinNumber', 'MinValue', 'Min', 'GetPotionType', 'SetItemName', 'ListLength', 'Sine', 'DirectionName', 'RepeatString', 'DuplicateText', 'RepeatText', 'GetItemLore', 'JoinString', 'JoinText', 'ReverseList', 'DedupList', 'CreateDict', 'GetBlockByMCTag', 'RoundNumber', 'FaceLocation', 'GetItemLoreLine', 'SetVectorLength', 'JsonFromValue', 'SetPotionDur', 'BlockResistance', 'SplitString', 'SplitText', 'NormalRandom', 'SetPotionType', 'AlignVector', 'SetItemDura', 'SetBreakability', 'SetMaxAmount', ' GetSignText ', 'RaycastEntity', 'SetDictValue', 'SetAllCoords', 'RGBColor', 'SetCanDestroy', 'HSLColor', ' GetDirection ', ' GetItemLore ', 'RemoveListIndex', 'CellularNoise', 'Base64Encode', 'Logarithm', 'SetItemTag', 'TrimString', 'TrimText', 'ParseMiniMessageExpr', 'GetItemAmount', 'SetPotionAmp', 'GetCanDestroy', 'RotateAroundAxis', 'GetItemName', 'GetItemDura', 'ShiftInDirection', 'WrapNum', 'ReplaceString', 'ReplaceText', 'SetItemGlowing', ' SetItemName ', 'SetLodestoneLoc', 'FlattenList', 'BlockHardness', 'GetPotionAmp', 'GetParticleAmount', 'GetDictSize', 'SetItemAmount', 'SubtractVectors', 'SetCase', 'SetParticleColor', 'GetLight', ' GetBookText ', 'GetDictValues', 'Vector', 'Distance', 'SetItemLore', 'ParseMiniMessage', 'Root', 'SquareRoot', 'SetParticleAmount', 'AddItemEnchant', 'AddItemToolRule', 'GetItemType', 'GetDirection', 'GetLoreLine', 'ArcTangent2', 'GetParticleType', 'SetItemMaxDura', 'RemoveString', 'RemoveText', 'GetAllBlockData', 'SetCrossbowProj', 'MaxNumber', 'MaxValue', 'Max', 'GetDictKeys', 'TrimStyledText', 'SetParticleMat', 'GetCoord', 'RemoveItemTag', 'SetParticleSize', 'GetPotionDur', 'RandomLoc', 'SetSoundType', 'GetLodestoneLoc', 'ShiftDirection', 'GetContainerName', 'GetParticleSprd', 'ReflectVector', 'ClearItemAttrs', 'GetHeadOwner', 'GetItemEnchants', 'AppendDict', 'GetMaxItemAmount', 'GetColorChannels', ' SetDirection ', 'SetListValue', ' SetItemEnchants ', 'SetBookText', 'RandomValue', 'RandomObj', ' SetItemFlags ', 'SetItemType', 'GetSoundType', 'GetListValue', 'ValueToJson', 'BounceNum', 'Tangent', 'VoronoiNoise', 'SetDirection', 'FaceDirection', 'HSBColor', 'HSVColor', '+=', 'Interpolate', 'GetSoundVariant', 'GetItemColor', 'ClearFormatting', 'InsertListValue', 'InsertListIndex', 'SetSoundVolume', 'SetCoord', 'AddVectors', 'SanitizeTags', 'SetPitch', 'GetParticleFade', 'RaycastBlock', 'SetItemTool', ' GetItemEnchants ', 'SetHeadTexture', 'SetHeadOwner', 'GzipCompress', 'PerlinNoise', 'WorleyNoise', 'SetItemColor', 'SetLeatherColor', 'GetParticleColor', 'SetSoundPitch', ' RoundNumber ', 'GetCanPlaceOn', 'SortList', 'SetCustomSound', 'RemoveDictEntry', 'FormatTime', 'SetItemFlags', 'StringLength', 'TextLength', 'GetItemEffects', 'StyledText', 'GetMiniMessageExpr', 'SetYaw', ' SetItemLore ', 'SetItemEffects', '-=', 'GetItemTag', 'CreateList', 'AppendList', 'GetContainerItems', 'ShiftToward', 'ShiftTowards', 'ShiftLocTowards', 'TrimList', 'GradientNoise', 'SetItemHideTooltip', 'GetBlockDrops', 'ClearEnchants', 'Cosine', 'GetParticleOpac', 'GetItemByMCTag', 'SetParticleFade', 'SetVectorComp', 'ParseNumber', 'Exponent', 'ShiftAllDirs', 'GetValueIndex', 'RemItemEnchant', 'AddItemLore', 'GetBookText', 'SetParticleRoll', 'SetSoundVariant', 'ShiftLocation', 'RandomizeList', 'ClampNumber', 'Round', 'GetSoundPitch', 'TranslateColors', 'GetBlockGrowth', 'GetAllItemTags', 'RemoveListValue', 'BytesToString', 'ShiftAllDirections', 'ValueNoise', 'WebResponse', 'SetCanPlaceOn', 'GetBlockType', 'ParsePitch', 'GetDictValue', 'ContainerLock', 'GetBlockPower', 'GetVectorLength', 'ContentLength', 'SetModelData', 'StringToBytes', 'SetCoords', 'GetMaxAmount', 'GetParticleSize']
IF_PLAYER_ACTION = Literal['HasRoomForItem', 'PHasRoomForItem', 'IsHoldingOff', 'NoItemCooldown', 'IsUsingItem', 'HasAllItems', 'HasItem', 'BlockEquals', 'IsWearing', 'PIsWearing', 'IsNear', 'PIsNear', 'IsRiding', 'CmdEquals', 'CursorItem', 'SlotEquals', 'ItemEquals', 'MenuSlotEquals', 'HasPermission', ' IsRiding ', 'MainHandEquals', 'IsSneaking', 'IsFlying', 'InvOpen', 'HasSlotItem', 'IsSprinting', 'IsGliding', 'CmdArgEquals', 'IsLookingAt', 'InWorldBorder', 'IsInGameMode', 'UsingPack', 'IsSwimming', 'StandingOn', 'PStandingOn', ' StandingOn ', ' PStandingOn ', 'IsGrounded', 'OwnsProduct', 'IsHoldingMain', 'IsHolding', 'IsBlocking', 'IsHitboxNear', 'PHitboxIsNear', 'HasPotion', 'HasEffect', 'NameEquals', 'PNameEquals', 'CanFly']
IF_ENTITY_ACTION = Literal['IsVehicle', 'IsGrounded', 'EIsGrounded', 'IsType', 'IsProj', 'IsMob', 'HasCustomTag', 'IsSheared', 'IsItem', ' IsRiding ', 'Exists', 'IsHitboxNear', 'EHitboxIsNear', 'IsNear', 'EIsNear', 'HasPotion', 'IsRiding', 'StandingOn', 'EStandingOn', 'NameEquals', 'ENameEquals', ' StandingOn ', ' EStandingOn ']
IF_GAME_ACTION = Literal['SignHasTxt', 'HasRoomForItem', 'EventBlockEquals', 'CommandEquals', 'EventItemEquals', ' SignHasTxt ', 'AttackIsCrit', 'ContainerHas', 'BlockEquals', 'GBlockEquals', 'InBlock', 'BlockPowered', 'HasPlayer', 'ContainerHasAll', 'CmdArgEquals', 'EventCancelled', 'IsChunkLoaded']
IF_VAR_ACTION = Literal['<=', 'ItemHasEnchant', 'ItemIsBlock', 'DictValueEquals', 'ItemHasTag', 'ListSizeEquals', 'ListIsEmpty', 'StringMatches', ' TextMatches ', 'StartsWith', 'ListValueEq', 'VarIsType', 'TextMatches', 'EqIgnoreCase', 'IsNear', 'VIsNear', ' InRange ', 'VarExists', 'BlockIsSolid', 'ItemEquals', 'VItemEquals', 'ListContains', 'InRange', 'LocIsNear', 'Contains', '!=', '<', '=', ' = ', '>', 'EndsWith', '>=', 'VarIsEmpty', 'DictHasKey']
REPEAT_ACTION = Literal['Adjacent', 'Path', 'Multiple', 'N Times', 'Grid', 'While', 'WhileCond', 'Range', 'ForEach', 'Sphere', 'Forever', ' Range ', 'ForEachEntry']
SELECT_OBJ_ACTION = Literal['LastMob', 'RandomPlayer', 'LastEntity', 'Shooter', 'AllMobs', 'EntityName', 'FilterRandom', 'RandomSelected', 'DefaultEntity', 'PlayerName', 'AllEntities', 'Damager', 'FilterDistance', 'FilterRay', 'Reset', 'None', 'EventTarget', 'Killer', 'Victim', 'EntitiesCond', 'AllPlayers', 'Invert', 'RandomEntity', 'FilterCondition', 'FilterSelect', 'MobsCond', 'FilterSort', 'Projectile', 'DefaultPlayer', 'PlayersCond', 'MobName']
CONTROL_ACTION = Literal['StopRepeat', 'Return', 'ReturnNTimes', 'Skip', 'End', 'Wait']
EVENT_ACTION = Literal['ClickContainerSlot', 'CloseInv', 'StartFly', 'BreakBlock', 'StartSprint', 'MobKillPlayer', 'Teleport', 'ShootBow', 'ChangeSign', 'StopFly', 'TameEntity', 'LeftClick', 'PlayerTakeDmg', 'ProjHit', 'KillPlayer', 'VehicleJump', 'ClickInvSlot', 'ClickOwnInv', 'Respawn', 'SwapHands', 'PackLoad', 'DamageEntity', 'Sneak', 'PlayerHeal', 'ClickPlayer', 'Consume', 'StartGlide', 'Death', 'PlaceBlock', 'Walk', 'PickUpItem', 'PickupItem', 'Dismount', 'CloudImbuePlayer', 'Leave', 'Quit', 'DropItem', 'LeftClickPlayer', 'ChangeSlot', 'ClickEntity', 'HorseJump', 'ShootProjectile', 'Move', 'Unsneak', 'PlayerCombust', 'Chat', 'Fish', 'FishEvent', 'FallDamage', 'BreakItem', 'LoopEvent', 'PlayerResurrect', 'RightClick', 'Purchase', 'ClickMenuSlot', 'ClickItem', 'Riptide', 'StopGlide', 'KillMob', 'Join', 'EntityDmgPlayer', 'StopSprint', 'Jump', 'LoadCrossbow', 'ProjDmgPlayer', 'Command', 'LeftClickEntity', 'Exhaustion', 'PackDecline', 'PlayerDmgPlayer']
ENTITY_EVENT_ACTION = Literal['EntityKillEntity', 'EntityCombust', 'BlockFall', 'ProjKillEntity', 'EntityDmgEntity', 'FallingBlockLand', 'EntityResurrect', 'ItemMerge', 'EntityHeal', 'Teleport', 'ShootBow', 'EntityDmg', 'ProjDmgEntity', 'EntityExplode', 'EntityDeath', 'VehicleDamage', 'Transform', 'RegrowWool']
GAME_VALUE_NAME = Literal['Current Health', 'Maximum Health', 'Absorption Health', 'Food Level', 'Food Saturation', 'Food Exhaustion', 'Attack Damage', 'Attack Speed', 'Attack Cooldown', 'Attack Cooldown Ticks', 'Armor Points', 'Armor Toughness', 'Invulnerability Ticks', 'Experience Level', 'Experience Progress', 'Fire Ticks', 'Freeze Ticks', 'Remaining Air', 'Fall Distance', 'Held Slot', 'Ping', 'Steer Sideways Movement', 'Steer Forward Movement', 'Item Usage Progress', 'Flight Speed', 'Walk Speed', 'Location', 'Target Block Location', 'Target Block Side', 'Eye Location', 'X-Coordinate', 'Y-Coordinate', 'Z-Coordinate', 'Midpoint Location', 'Pitch', 'Yaw', 'Body Yaw', 'Standing Block Location', 'Spawn Location', 'Velocity', 'Direction', 'Entity Width', 'Entity Height', 'Main Hand Item', 'Off Hand Item', 'Armor Items', 'Hotbar Items', 'Inventory Items', 'Cursor Item', 'Inventory Menu Items', 'Saddle Item', 'Entity Item', 'Name ', 'UUID', 'Entity Type', 'Game Mode', 'Open Inventory Title ', 'Potion Effects', 'Vehicle', 'Passengers ', 'Lead Holder', 'Attached Leads', 'Targeted Entity UUID', 'Projectile Shooter UUID', 'Display Entity Translation', 'Display Entity Scale', 'Pose', 'Event Block Location', 'Event Block Side', 'Event Damage', 'Damage Event Cause', 'Raw Event Damage', 'Event Death Message', 'Event Heal Amount', 'Heal Event Cause', 'Event Explosion Affected Blocks', 'Event Power', 'Event Command', 'Event Command Arguments', 'Event Item', 'Event Hotbar Slot', 'Event Clicked Slot Index', 'Event Clicked Slot Item', 'Event Clicked Slot New Item', 'Close Inventory Event Cause', 'Inventory Event Click Type', 'Fish Event Cause', 'Teleport Event Cause', 'Teleport Location', 'Exhaustion Event Cause', 'Event Exhaustion', 'Transform Event Cause', 'Event Transform Entities', 'Event Hit Type', 'Product ID', 'Event Message', 'Event Sign Text', 'Event Sign Side', 'Combust Event Duration', 'Combust Event Cause', 'Player Count', 'CPU Usage', 'Server TPS', 'Timestamp', 'Selection Size', 'Selection Target Names', 'Selection Target UUIDs', 'Plot ID', 'Plot Name', 'Plot Size', 'Microseconds Since Startup', 'Ticks Since Startup', 'Active Block Transactions', 'Plot Player Names', 'Plot Player UUIDs']
SOUND_NAME = Literal['Axolotl Attack', 'Axolotl Death', 'Axolotl Hurt', 'Axolotl Ambient Air', 'Axolotl Ambient Water', 'Axolotl Splash', 'Axolotl Swim', 'Bat Ambient', 'Bat Death', 'Bat Fly', 'Bat Hurt', 'Bat Takeoff', 'Bee Death', 'Bee Hurt', 'Bee Loop', 'Bee Loop (Aggressive)', 'Bee Pollinate', 'Bee Sting', 'Camel Ambient', 'Camel Dash Ready', 'Camel Dash', 'Camel Death', 'Camel Eat', 'Camel Hurt', 'Camel Equip Saddle', 'Camel Sit', 'Camel Stand', 'Camel Step', 'Camel Step on Sand', 'Cat Ambient', 'Cat Beg For Food', 'Cat Death', 'Cat Eat', 'Cat Hiss', 'Cat Hurt', 'Cat Purr', 'Cat Purreow', 'Stray Cat Ambient', 'Chicken Ambient', 'Chicken Death', 'Chicken Hurt', 'Chicken Lay Egg', 'Chicken Step', 'Cow Ambient', 'Cow Death', 'Cow Hurt', 'Cow Milk', 'Cow Step', 'Dolphin Ambient', 'Dolphin Ambient (Water)', 'Dolphin Attack', 'Dolphin Death', 'Dolphin Eat', 'Dolphin Hurt', 'Dolphin Jump', 'Dolphin Play', 'Dolphin Splash', 'Dolphin Swim', 'Donkey Ambient', 'Donkey Angry', 'Donkey Equip Chest', 'Donkey Death', 'Donkey Eat', 'Donkey Hurt', 'Donkey Jump', 'Fish Swim', 'Cod Ambient', 'Cod Death', 'Cod Flop', 'Cod Hurt', 'Salmon Ambient', 'Salmon Death', 'Salmon Flop', 'Salmon Hurt', 'Tropical Fish Ambient', 'Tropical Fish Death', 'Tropical Fish Flop', 'Tropical Fish Hurt', 'Fox Aggravate', 'Fox Ambient', 'Fox Bite', 'Fox Death', 'Fox Eat', 'Fox Hurt', 'Fox Screech', 'Fox Sleep', 'Fox Sniff', 'Fox Spit', 'Fox Teleport', 'Goat Ambient', 'Goat Death', 'Goat Eat', 'Goat Hurt', 'Goat Long Jump', 'Goat Milk', 'Goat Prepare Ram', 'Goat Ram Impact', 'Goat Horn Break', 'Goat Step', 'Screaming Goat Ambient', 'Screaming Goat Death', 'Screaming Goat Eat', 'Screaming Goat Hurt', 'Screaming Goat Long Jump', 'Screaming Goat Milk', 'Screaming Goat Prepare Ram', 'Screaming Goat Ram Impact', 'Glow Squid Ambient', 'Glow Squid Death', 'Glow Squid Hurt', 'Glow Squid Squirt', 'Horse Ambient', 'Horse Angry', 'Horse Breathe', 'Horse Death', 'Horse Eat', 'Horse Equip Saddle', 'Horse Equip Armor', 'Horse Gallop', 'Horse Hurt', 'Horse Jump', 'Horse Land', 'Horse Step', 'Horse Step (Wood)', 'Iron Golem Attack', 'Iron Golem Damage', 'Iron Golem Death', 'Iron Golem Hurt', 'Iron Golem Repair', 'Iron Golem Step', 'Llama Ambient', 'Llama Angry', 'Llama Death', 'Llama Eat', 'Llama Equip Chest', 'Llama Hurt', 'Llama Step', 'Llama Spit', 'Llama Swag', 'Mooshroom Convert', 'Mooshroom Eat', 'Mooshroom Milk', 'Mooshroom Shear', 'Brown Mooshroom Milk', 'Mule Angry', 'Mule Equip Chest', 'Mule Ambient', 'Mule Death', 'Mule Eat', 'Mule Hurt', 'Mule Jump', 'Ocelot Ambient', 'Ocelot Death', 'Ocelot Hurt', 'Panda Ambient', 'Panda Ambient (Aggressive)', 'Panda Ambient (Worried)', 'Panda Bite', 'Panda Sterile', 'Panda Death', 'Panda Eat', 'Panda Hurt', 'Panda Pre-Sneeze', 'Panda Sneeze', 'Panda Step', 'Parrot Ambient', 'Parrot Death', 'Parrot Hurt', 'Parrot Step', 'Parrot Eat', 'Parrot Fly', 'Parrot Imitate Blaze', 'Parrot Imitate Bogged', 'Parrot Imitate Breeze', 'Parrot Imitate Creeper', 'Parrot Imitate Drowned', 'Parrot Imitate Elder Guardian', 'Parrot Imitate Ender Dragon', 'Parrot Imitate Endermite', 'Parrot Imitate Evoker', 'Parrot Imitate Ghast', 'Parrot Imitate Guardian', 'Parrot Imitate Hoglin', 'Parrot Imitate Husk', 'Parrot Imitate Illusioner', 'Parrot Imitate Magma Cube', 'Parrot Imitate Phantom', 'Parrot Imitate Piglin', 'Parrot Imitate Pillager', 'Parrot Imitate Ravager', 'Parrot Imitate Piglin Brute', 'Parrot Imitate Shulker', 'Parrot Imitate Silverfish', 'Parrot Imitate Skeleton', 'Parrot Imitate Slime', 'Parrot Imitate Spider', 'Parrot Imitate Stray', 'Parrot Imitate Vex', 'Parrot Imitate Vindicator', 'Parrot Imitate Warden', 'Parrot Imitate Witch', 'Parrot Imitate Wither', 'Parrot Imitate Wither Skeleton', 'Parrot Imitate Zoglin', 'Parrot Imitate Zombie', 'Parrot Imitate Zombie Villager', 'Pig Ambient', 'Pig Death', 'Pig Equip Saddle', 'Pig Hurt', 'Pig Step', 'Polar Bear Ambient', 'Polar Bear Death', 'Polar Bear Hurt', 'Polar Bear Step', 'Polar Bear Warn', 'Baby Polar Bear Ambient', 'Pufferfish Ambient', 'Pufferfish Death', 'Pufferfish Deflate', 'Pufferfish Flop', 'Pufferfish Hurt', 'Pufferfish Inflate', 'Pufferfish Sting', 'Rabbit Ambient', 'Rabbit Death', 'Rabbit Hurt', 'Rabbit Jump', 'Killer Rabbit Attack', 'Sheep Ambient', 'Sheep Death', 'Sheep Hurt', 'Sheep Shear', 'Sheep Step', 'Sniffer Ambient', 'Sniffer Death', 'Sniffer Digging', 'Sniffer Stop Digging', 'Sniffer Drop Seed', 'Sniffer Eat', 'Sniffer Happy', 'Sniffer Hurt', 'Sniffer Scenting', 'Sniffer Searching', 'Sniffer Sniffing', 'Sniffer Step', 'Snow Golem Ambient', 'Snow Golem Death', 'Snow Golem Hurt', 'Snow Golem Shear', 'Snow Golem Shoot', 'Squid Ambient', 'Squid Death', 'Squid Hurt', 'Squid Squirt', 'Tadpole Death', 'Tadpole Hurt', 'Tadpole Flop', 'Tadpole Grow Up', 'Turtle Ambient', 'Turtle Death', 'Turtle Hurt', 'Turtle Shamble', 'Turtle Lay Egg', 'Turtle Egg Crack', 'Turtle Egg Hatch', 'Turtle Egg Break', 'Turtle Swim', 'Baby Turtle Death', 'Baby Turtle Hurt', 'Baby Turtle Shamble', 'Wolf Ambient', 'Wolf Death', 'Wolf Growl', 'Wolf Howl', 'Wolf Hurt', 'Wolf Pant', 'Wolf Shake', 'Wolf Step', 'Wolf Whine', 'Allay Ambient (Without Item)', 'Allay Ambient (With Item)', 'Allay Death', 'Allay Hurt', 'Allay Item Given', 'Allay Item Pickup', 'Allay Item Thrown', 'Armadillo Ambient', 'Armadillo Brush', 'Armadillo Death', 'Armadillo Eat', 'Armadillo Hurt', 'Armadillo Hurt (Reduced)', 'Armadillo Land', 'Armadillo Peek', 'Armadillo Roll', 'Armadillo Scute Drop', 'Armadillo Step', 'Armadillo Unroll Finish', 'Armadillo Unroll Start', 'Bogged Ambient', 'Bogged Death', 'Bogged Hurt', 'Bogged Shear', 'Bogged Step', 'Breeze Charge', 'Breeze Death', 'Breeze Deflect', 'Breeze Hurt', 'Breeze Idle (Air)', 'Breeze Idle (Ground)', 'Breeze Inhale', 'Breeze Jump', 'Breeze Land', 'Breeze Shoot', 'Breeze Slide', 'Breeze Whirl', 'Breeze Wind Burst', 'Frog Ambient', 'Frog Death', 'Frog Eat', 'Frog Hurt', 'Frog Step', 'Frog Lay Spawn', 'Frog Long Jump', 'Frog Tongue', 'Strider Ambient', 'Strider Death', 'Strider Eat', 'Strider Equip Saddle', 'Strider Happy', 'Strider Hurt', 'Strider Retreat', 'Strider Step', 'Strider Step (Lava)', 'Villager Ambient', 'Villager Death', 'Villager Hurt', 'Villager No', 'Villager Trading', 'Villager Yes', 'Armorer Work', 'Butcher Work', 'Cartographer Work', 'Cleric Work', 'Farmer Work', 'Fisherman Work', 'Fletcher Work', 'Leatherworker Work', 'Librarian Work', 'Mason Work', 'Shepherd Work', 'Toolsmith Work', 'Weaponsmith Work', 'Villager Celebrate', 'Wandering Trader Ambient', 'Wandering Trader Death', 'Wandering Trader Disappear', 'Wandering Trader Drink Milk', 'Wandering Trader Drink Potion', 'Wandering Trader Hurt', 'Wandering Trader No', 'Wandering Trader Reappear', 'Wandering Trader Trading', 'Wandering Trader Yes', 'Creeper Death', 'Explode', 'Creeper Hurt', 'Creeper Primed', 'Drowned Ambient', 'Drowned Ambient (Land)', 'Drowned Death', 'Drowned Death (Land)', 'Drowned Hurt', 'Drowned Hurt (Land)', 'Drowned Shoot', 'Drowned Step', 'Drowned Swim', 'Elder Guardian Ambient', 'Elder Guardian Ambient (Land)', 'Elder Guardian Curse', 'Elder Guardian Death', 'Elder Guardian Death (Land)', 'Elder Guardian Flop', 'Elder Guardian Hurt', 'Elder Guardian Hurt (Land)', 'Guardian Ambient', 'Guardian Ambient (Land)', 'Guardian Attack', 'Guardian Death', 'Guardian Death (Land)', 'Guardian Flop', 'Guardian Hurt', 'Guardian Hurt (Land)', 'Husk Ambient', 'Husk Death', 'Husk Hurt', 'Husk Step', 'Husk Drown', 'Phantom Ambient', 'Phantom Bite', 'Phantom Death', 'Phantom Flap', 'Phantom Hurt', 'Phantom Swoop', 'Ravager Ambient', 'Ravager Attack', 'Ravager Death', 'Ravager Hurt', 'Ravager Roar', 'Ravager Step', 'Ravager Stunned', 'Ravager Celebrate', 'Silverfish Ambient', 'Silverfish Death', 'Silverfish Hurt', 'Silverfish Step', 'Skeleton Ambient', 'Skeleton Convert To Stray', 'Skeleton Death', 'Skeleton Hurt', 'Skeleton Step', 'Skeleton Shoot', 'Skeleton Horse Ambient', 'Skeleton Horse Ambient (Water)', 'Skeleton Horse Death', 'Skeleton Horse Gallop (Water)', 'Skeleton Horse Hurt', 'Skeleton Horse Jump (Water)', 'Skeleton Horse Step (Water)', 'Skeleton Horse Swim', 'Slime Attack', 'Slime Death', 'Slime Hurt', 'Slime Jump', 'Slime Squish', 'Small Slime Death', 'Small Slime Hurt', 'Small Slime Jump', 'Small Slime Squish', 'Spider Ambient', 'Spider Death', 'Spider Hurt', 'Spider Step', 'Stray Ambient', 'Stray Death', 'Stray Hurt', 'Stray Step', 'Vex Ambient', 'Vex Charge', 'Vex Death', 'Vex Hurt', 'Warden Ambient', 'Warden Death', 'Warden Hurt', 'Warden Step', 'Warden Groan', 'Warden Angry', 'Warden Attack', 'Warden Dig', 'Warden Emerge', 'Warden Heartbeat', 'Warden Listening', 'Warden Listening (Angry)', 'Warden Approach', 'Warden Approach (Closer)', 'Warden Approach (Closest)', 'Warden Roar', 'Warden Sniff', 'Warden Sonic Boom', 'Warden Sonic Charge', 'Warden Tendril Click', 'Witch Ambient', 'Witch Death', 'Witch Drink', 'Witch Hurt', 'Witch Throw', 'Witch Celebrate', 'Zombie Ambient', 'Zombie Attack Wooden Door', 'Zombie Attack Iron Door', 'Zombie Break Wooden Door', 'Zombie Death', 'Zombie Destroy Egg', 'Zombie Drown', 'Zombie Infect', 'Zombie Hurt', 'Zombie Step', 'Zombie Horse Ambient', 'Zombie Horse Death', 'Zombie Horse Hurt', 'Zombie Villager Ambient', 'Zombie Villager Convert', 'Zombie Villager Cure', 'Zombie Villager Death', 'Zombie Villager Hurt', 'Zombie Villager Step', 'Blaze Ambient', 'Blaze Burn', 'Blaze Death', 'Blaze Hurt', 'Blaze Shoot', 'Ghast Ambient', 'Ghast Death', 'Ghast Hurt', 'Ghast Scream', 'Ghast Shoot', 'Ghast Warn', 'Hoglin Ambient', 'Hoglin Angry', 'Hoglin Attack', 'Hoglin Zombify', 'Hoglin Death', 'Hoglin Hurt', 'Hoglin Retreat', 'Hoglin Step', 'Magma Cube Death', 'Magma Cube Hurt', 'Magma Cube Jump', 'Magma Cube Squish', 'Small Magma Cube Death', 'Small Magma Cube Hurt', 'Small Magma Cube Squish', 'Piglin Admiring Item', 'Piglin Ambient', 'Piglin Angry', 'Piglin Celebrate', 'Piglin Zombify', 'Piglin Death', 'Piglin Hurt', 'Piglin Jealous', 'Piglin Retreat', 'Piglin Step', 'Piglin Brute Ambient', 'Piglin Brute Angry', 'Piglin Brute Zombify', 'Piglin Brute Death', 'Piglin Brute Hurt', 'Piglin Brute Step', 'Wither Ambient', 'Wither Break Block', 'Wither Death', 'Wither Hurt', 'Wither Shoot', 'Wither Spawn', 'Wither Skeleton Ambient', 'Wither Skeleton Death', 'Wither Skeleton Hurt', 'Wither Skeleton Step', 'Zoglin Ambient', 'Zoglin Angry', 'Zoglin Attack', 'Zoglin Death', 'Zoglin Hurt', 'Zoglin Step', 'Zombified Piglin Ambient', 'Zombified Piglin Angry', 'Zombified Piglin Death', 'Zombified Piglin Hurt', 'Enderman Ambient', 'Enderman Death', 'Enderman Hurt', 'Enderman Scream', 'Enderman Stare', 'Enderman Teleport', 'Endermite Ambient', 'Endermite Death', 'Endermite Hurt', 'Endermite Step', 'Ender Dragon Ambient', 'Ender Dragon Death', 'Ender Dragon Hurt', 'Ender Dragon Fireball Explode', 'Ender Dragon Flap', 'Ender Dragon Growl', 'Ender Dragon Shoot', 'Shulker Ambient', 'Shulker Close', 'Shulker Death', 'Shulker Hurt', 'Shulker Hurt (Closed)', 'Shulker Open', 'Shulker Shoot', 'Shulker Teleport', 'Shulker Bullet Explode', 'Shulker Bullet Break', 'Evoker Ambient', 'Evoker Cast Spell', 'Evoker Death', 'Evoker Hurt', 'Evoker Prepare Attack', 'Evoker Prepare Summon', 'Evoker Prepare Wololo', 'Evoker Fangs Attack', 'Evoker Celebrate', 'Illusioner Ambient', 'Illusioner Death', 'Illusioner Hurt', 'Illusioner Cast Spell', 'Illusioner Mirror Move', 'Illusioner Prepare Blindness', 'Illusioner Prepare Mirror', 'Pillager Ambient', 'Pillager Death', 'Pillager Hurt', 'Pillager Celebrate', 'Vindicator Ambient', 'Vindicator Death', 'Vindicator Hurt', 'Vindicator Celebrate', 'Player Attack (Critical)', 'Player Attack (Knockback)', 'Player Attack (No Damage)', 'Player Attack (Strong)', 'Player Attack (Sweep)', 'Player Attack (Weak)', 'Player Breathe', 'Player Burn', 'Player Freeze', 'Player Burp', 'Player Drown', 'Player Hurt (Sweet Berry Bush)', 'Player Level Up', 'Player Splash (High Speed)', 'Death', 'Drink', 'Eat', 'Extinguish', 'Fall', 'Fall (Big)', 'Hurt', 'Splash', 'Swim', 'Challenge Complete', 'UI Button Click', 'UI Toast In', 'UI Toast Out', 'Arrow Hit', 'Arrow Hit Player', 'Arrow Shoot', 'Egg Throw', 'Ender Eye Death', 'Ender Eye Launch', 'Enderpearl Throw', 'Experience Bottle Throw', 'Lingering Potion Throw', 'Splash Potion Break', 'Splash Potion Throw', 'Snowball Throw', 'Firework Blast', 'Firework Blast Far', 'Firework Large Blast', 'Firework Large Blast Far', 'Firework Launch', 'Firework Shoot', 'Firework Twinkle', 'Firework Twinkle Far', 'Glow Item Frame Add Item', 'Glow Item Frame Break', 'Glow Item Frame Place', 'Glow Item Frame Remove Item', 'Glow Item Frame Rotate Item', 'Item Frame Add Item', 'Item Frame Break', 'Item Frame Place', 'Item Frame Remove Item', 'Item Frame Rotate Item', 'Painting Break', 'Painting Place', 'Boat Paddle Land', 'Boat Paddle Water', 'Minecart Inside', 'Minecart Inside Water', 'Rolling Minecart', 'Armor Stand Break', 'Armor Stand Fall', 'Armor Stand Hit', 'Armor Stand Place', 'Burn', 'Experience Orb Pickup', 'Leash Knot Break', 'Leash Knot Place', 'TNT Primed', 'Bad Omen Mob Effect', 'Raid Omen Mob Effect', 'Trial Omen Mob Effect', 'Creative Music', 'Credits Music', 'Boss Music', 'End Music', 'Menu Music', 'Overworld Music', 'Dripstone Caves Music', 'Frozen Peaks Music', 'Grove Music', 'Jagged Peaks Music', 'Lush Caves Music', 'Meadow Music', 'Snowy Slopes Music', 'Stony Peaks Music', 'Forest Music', 'Jungle Music', 'Sparse Jungle Music', 'Bamboo Jungle Music', 'Old Growth Taiga Music', 'Swamp Music', 'Desert Music', 'Badlands Music', 'Flower Forest Music', 'Cherry Grove Music', 'Deep Dark Music', 'Basalt Deltas Music', 'Crimson Forest Music', 'Nether Wastes Music', 'Soul Sand Valley Music', 'Underwater Music', 'Music Disc - 5', 'Music Disc - 11', 'Music Disc - 13', 'Music Disc - Blocks', 'Music Disc - Cat', 'Music Disc - Chirp', 'Music Disc - Creator', 'Music Disc - Creator (Music Box)', 'Music Disc - Far', 'Music Disc - Mall', 'Music Disc - Mellohi', 'Music Disc - Pigstep', 'Music Disc - Stal', 'Music Disc - Strad', 'Music Disc - Ward', 'Music Disc - Wait', 'Music Disc - Otherside', 'Music Disc - Precipice', 'Music Disc - Relic', 'Shield Block', 'Shield Break', 'Crossbow Hit', 'Crossbow Loading (End)', 'Crossbow Loading (Middle)', 'Crossbow Loading (Start)', 'Crossbow Quick Charge (I)', 'Crossbow Quick Charge (II)', 'Crossbow Quick Charge (III)', 'Crossbow Shoot', 'Mace Smash Air', 'Mace Smash Ground', 'Mace Smash Ground (Heavy)', 'Armor Equip', 'Armor Equip (Elytra)', 'Armor Equip (Leather)', 'Armor Equip (Chain)', 'Armor Equip (Iron)', 'Armor Equip (Gold)', 'Armor Equip (Diamond)', 'Armor Equip (Netherite)', 'Armor Equip (Turtle)', 'Elytra Flying', 'Thorns Hit', 'Wolf Armor Equip', 'Wolf Armor Unequip', 'Wolf Armor Break', 'Wolf Armor Crack', 'Wolf Armor Damage', 'Wolf Armor Repair', 'Trident Throw', 'Trident Hit', 'Trident Hit Ground', 'Trident Return', 'Trident Riptide (I)', 'Trident Riptide (II)', 'Trident Riptide (III)', 'Trident Thunder', 'Fire Charge Use', 'Flint and Steel Use', 'Axe Scrape', 'Axe Strip', 'Wax Off', 'Hoe Till', 'Shovel Flatten', 'Fishing Bobber Retrieve', 'Fishing Bobber Splash', 'Fishing Bobber Throw', 'Brushing (Generic)', 'Brushing Gravel', 'Brushing Gravel Complete', 'Brushing Sand', 'Brushing Sand Complete', 'Bottle Empty', 'Bottle Fill', 'Dragon Breath Fill', 'Honey Bottle Drink', 'Water Bucket Empty', 'Water Bucket Fill', 'Axolotl Bucket Empty', 'Axolotl Bucket Fill', 'Fish Bucket Empty', 'Fish Bucket Fill', 'Tadpole Bucket Empty', 'Tadpole Bucket Fill', 'Lava Bucket Empty', 'Lava Bucket Fill', 'Powder Bucket Empty', 'Powder Bucket Fill', 'Item Break', 'Item Pickup', 'Book Page Turn', 'Book Put in Lectern', 'Bone Meal Use', 'Chorus Fruit Teleport', 'Dye Use', 'Ink Sac Use', 'Glow Ink Sac Use', 'Goat Horn (Ponder)', 'Goat Horn (Sing)', 'Goat Horn (Seek)', 'Goat Horn (Feel)', 'Goat Horn (Admire)', 'Goat Horn (Call)', 'Goat Horn (Yearn)', 'Goat Horn (Dream)', 'Spyglass Use', 'Spyglass Stop Using', 'Bundle Drop Contents', 'Bundle Remove One', 'Bundle Item Insert', 'Totem Use', 'Wax Copper', 'Wind Charge Throw', 'Wind Charge Burst', 'Ominous Bottle Dispose', 'Bamboo Break', 'Bamboo Fall', 'Bamboo Hit', 'Bamboo Place', 'Bamboo Sapling Break', 'Bamboo Sapling Hit', 'Bamboo Sapling Place', 'Bamboo Step', 'Bamboo Planks Break', 'Bamboo Planks Fall', 'Bamboo Planks Hit', 'Bamboo Planks Place', 'Bamboo Planks Step', 'Beehive Drip', 'Beehive Enter', 'Beehive Exit', 'Beehive Shear', 'Beehive Work', 'Bone Block Break', 'Bone Block Fall', 'Bone Block Hit', 'Bone Block Place', 'Bone Block Step', 'Bubble Column Bubble Pop', 'Bubble Column Ambient (Up)', 'Bubble Column Inside (Up)', 'Bubble Column Ambient (Down)', 'Bubble Column Inside (Down)', 'Cherry Leaves Break', 'Cherry Leaves Fall', 'Cherry Leaves Hit', 'Cherry Leaves Place', 'Cherry Leaves Step', 'Cherry Sapling Break', 'Cherry Sapling Fall', 'Cherry Sapling Hit', 'Cherry Sapling Place', 'Cherry Sapling Step', 'Cherry Wood Break', 'Cherry Wood Fall', 'Cherry Wood Hit', 'Cherry Wood Place', 'Cherry Wood Step', 'Cobweb Break', 'Cobweb Fall', 'Cobweb Hit', 'Cobweb Place', 'Cobweb Step', 'Copper Bulb Break', 'Copper Bulb Fall', 'Copper Bulb Hit', 'Copper Bulb Place', 'Copper Bulb Step', 'Copper Bulb Turn Off', 'Copper Bulb Turn On', 'Copper Door Close', 'Copper Door Open', 'Copper Grate Break', 'Copper Grate Fall', 'Copper Grate Hit', 'Copper Grate Place', 'Copper Grate Step', 'Copper Trapdoor Close', 'Copper Trapdoor Open', 'Crafter Craft', 'Crafter Fail', 'Coral Block Break', 'Coral Block Fall', 'Coral Block Hit', 'Coral Block Place', 'Coral Block Step', 'Crop Plant', 'Crop Break', 'Plant Stop Growing', 'Fire Ambient', 'Fire Extinguish', 'Grass Break', 'Grass Fall', 'Grass Hit', 'Grass Place', 'Grass Step', 'Gravel Break', 'Gravel Fall', 'Gravel Hit', 'Gravel Place', 'Gravel Step', 'Lava Ambient', 'Lava Extinguish', 'Lava Pop', 'Lily Pad Place', 'Pumpkin Carve', 'Pink Petals Break', 'Pink Petals Fall', 'Pink Petals Hit', 'Pink Petals Place', 'Pink Petals Step', 'Powder Snow Break', 'Powder Snow Fall', 'Powder Snow Hit', 'Powder Snow Place', 'Powder Snow Step', 'Rooted Dirt Break', 'Rooted Dirt Fall', 'Rooted Dirt Hit', 'Rooted Dirt Place', 'Rooted Dirt Step', 'Sand Break', 'Sand Fall', 'Sand Hit', 'Sand Place', 'Sand Step', 'Suspicious Sand Break', 'Suspicious Sand Fall', 'Suspicious Sand Hit', 'Suspicious Sand Place', 'Suspicious Sand Step', 'Suspicious Gravel Break', 'Suspicious Gravel Fall', 'Suspicious Gravel Hit', 'Suspicious Gravel Place', 'Suspicious Gravel Step', 'Sniffer Egg Plop', 'Sniffer Egg Crack', 'Sniffer Egg Hatch', 'Snow Break', 'Snow Fall', 'Snow Hit', 'Snow Place', 'Snow Step', 'Stone Break', 'Stone Fall', 'Stone Hit', 'Stone Place', 'Stone Step', 'Sweet Berry Bush Break', 'Sweet Berry Bush Pick', 'Sweet Berry Bush Place', 'Seagrass Break', 'Seagrass Fall', 'Seagrass Hit', 'Seagrass Place', 'Seagrass Step', 'Vine Break', 'Vine Fall', 'Vine Hit', 'Vine Place', 'Vine Climb', 'Water Ambient', 'Water Enter', 'Water Exit', 'Wood Break', 'Wood Fall', 'Wood Hit', 'Wood Place', 'Wood Step', 'Nether Wood Break', 'Nether Wood Fall', 'Nether Wood Hit', 'Nether Wood Place', 'Nether Wood Step', 'Froglight Break', 'Froglight Fall', 'Froglight Hit', 'Froglight Place', 'Froglight Step', 'Frogspawn Break', 'Frogspawn Fall', 'Frogspawn Hit', 'Frogspawn Place', 'Frogspawn Step', 'Frogspawn Hatch', 'Mangrove Roots Break', 'Mangrove Roots Fall', 'Mangrove Roots Hit', 'Mangrove Roots Place', 'Mangrove Roots Step', 'Muddy Mangrove Roots Break', 'Muddy Mangrove Roots Fall', 'Muddy Mangrove Roots Hit', 'Muddy Mangrove Roots Place', 'Muddy Mangrove Roots Step', 'Mud Break', 'Mud Fall', 'Mud Hit', 'Mud Place', 'Mud Step', 'Amethyst Block Break', 'Amethyst Block Chime', 'Amethyst Block Fall', 'Amethyst Block Hit', 'Amethyst Block Place', 'Amethyst Block Step', 'Amethyst Cluster Break', 'Amethyst Cluster Fall', 'Amethyst Cluster Hit', 'Amethyst Cluster Place', 'Amethyst Cluster Step', 'Large Amethyst Bud Break', 'Large Amethyst Bud Place', 'Medium Amethyst Bud Break', 'Medium Amethyst Bud Place', 'Small Amethyst Bud Break', 'Small Amethyst Bud Place', 'Amethyst Resonate', 'Azalea Place', 'Azalea Step', 'Azalea Break', 'Azalea Fall', 'Azalea Hit', 'Azalea Leaves Break', 'Azalea Leaves Fall', 'Azalea Leaves Hit', 'Azalea Leaves Place', 'Azalea Leaves Step', 'Big Dripleaf Break', 'Big Dripleaf Fall', 'Big Dripleaf Hit', 'Big Dripleaf Break', 'Big Dripleaf Step', 'Big Dripleaf Tilt Down', 'Big Dripleaf Tilt Up', 'Calcite Break', 'Calcite Fall', 'Calcite Hit', 'Calcite Place', 'Calcite Step', 'Cave Vines Break', 'Cave Vines Fall', 'Cave Vines Hit', 'Cave Vines Pick', 'Cave Vines Place', 'Cave Vines Step', 'Deepslate Break', 'Deepslate Fall', 'Deepslate Hit', 'Deepslate Place', 'Deepslate Step', 'Deepslate Bricks Break', 'Deepslate Bricks Fall', 'Deepslate Bricks Hit', 'Deepslate Bricks Place', 'Deepslate Bricks Step', 'Deepslate Tiles Break', 'Deepslate Tiles Fall', 'Deepslate Tiles Hit', 'Deepslate Tiles Place', 'Deepslate Tiles Step', 'Dripstone Block Break', 'Dripstone Block Fall', 'Dripstone Block Hit', 'Dripstone Block Place', 'Dripstone Block Step', 'Flowering Azalea Break', 'Flowering Azalea Fall', 'Flowering Azalea Hit', 'Flowering Azalea Place', 'Flowering Azalea Step', 'Hanging Roots Break', 'Hanging Roots Fall', 'Hanging Roots Hit', 'Hanging Roots Place', 'Hanging Roots Step', 'Moss Block Break', 'Moss Block Fall', 'Moss Block Hit', 'Moss Block Place', 'Moss Block Step', 'Moss Carpet Break', 'Moss Carpet Fall', 'Moss Carpet Hit', 'Moss Carpet Place', 'Moss Carpet Step', 'Pointed Dripstone Break', 'Pointed Dripstone Drip Lava', 'Pointed Dripstone Drip Lava Into Cauldron', 'Pointed Dripstone Drip Water', 'Pointed Dripstone Drip Water Into Cauldron', 'Pointed Dripstone Fall', 'Pointed Dripstone Hit', 'Pointed Dripstone Land', 'Pointed Dripstone Place', 'Pointed Dripstone Step', 'Small Dripleaf Break', 'Small Dripleaf Fall', 'Small Dripleaf Hit', 'Small Dripleaf Place', 'Small Dripleaf Step', 'Sponge Absorb', 'Sponge Break', 'Sponge Fall', 'Sponge Hit', 'Sponge Place', 'Sponge Step', 'Wet Sponge Break', 'Wet Sponge Dries', 'Wet Sponge Fall', 'Wet Sponge Hit', 'Wet Sponge Place', 'Wet Sponge Step', 'Spore Blossom Break', 'Spore Blossom Fall', 'Spore Blossom Hit', 'Spore Blossom Place', 'Spore Blossom Step', 'Tuff Break', 'Tuff Fall', 'Tuff Hit', 'Tuff Place', 'Tuff Step', 'Sculk Break', 'Sculk Fall', 'Sculk Hit', 'Sculk Place', 'Sculk Step', 'Sculk Charge', 'Sculk Spread', 'Sculk Catalyst Break', 'Sculk Catalyst Fall', 'Sculk Catalyst Hit', 'Sculk Catalyst Place', 'Sculk Catalyst Step', 'Sculk Catalyst Bloom', 'Sculk Shrieker Break', 'Sculk Shrieker Fall', 'Sculk Shrieker Hit', 'Sculk Shrieker Place', 'Sculk Shrieker Step', 'Sculk Shrieker Shriek', 'Sculk Vein Break', 'Sculk Vein Fall', 'Sculk Vein Hit', 'Sculk Vein Place', 'Sculk Vein Step', 'Ancient Debris Break', 'Ancient Debris Fall', 'Ancient Debris Hit', 'Ancient Debris Place', 'Ancient Debris Step', 'Basalt Break', 'Basalt Fall', 'Basalt Hit', 'Basalt Place', 'Basalt Step', 'Gilded Blackstone Break', 'Gilded Blackstone Fall', 'Gilded Blackstone Hit', 'Gilded Blackstone Place', 'Gilded Blackstone Step', 'Netherrack Break', 'Netherrack Fall', 'Netherrack Hit', 'Netherrack Place', 'Netherrack Step', 'Nether Gold Ore Break', 'Nether Gold Ore Fall', 'Nether Gold Ore Hit', 'Nether Gold Ore Place', 'Nether Gold Ore Step', 'Nether Fungus Break', 'Nether Fungus Fall', 'Nether Fungus Hit', 'Nether Fungus Place', 'Nether Fungus Step', 'Nether Quartz Ore Break', 'Nether Quartz Ore Fall', 'Nether Quartz Ore Hit', 'Nether Quartz Ore Place', 'Nether Quartz Ore Step', 'Nether Roots Break', 'Nether Roots Fall', 'Nether Roots Hit', 'Nether Roots Place', 'Nether Roots Step', 'Nether Sprouts Break', 'Nether Sprouts Fall', 'Nether Sprouts Hit', 'Nether Sprouts Place', 'Nether Sprouts Step', 'Nether Stem Break', 'Nether Stem Fall', 'Nether Stem Hit', 'Nether Stem Place', 'Nether Stem Step', 'Nether Vine Break', 'Nether Vine Fall', 'Nether Vine Hit', 'Nether Vine Place', 'Nether Vine Step', 'Nether Wart Break', 'Nether Wart Plant', 'Nether Wart Block Break', 'Nether Wart Block Fall', 'Nether Wart Block Hit', 'Nether Wart Block Place', 'Nether Wart Block Step', 'Nylium Break', 'Nylium Fall', 'Nylium Hit', 'Nylium Place', 'Nylium Step', 'Shroomlight Break', 'Shroomlight Fall', 'Shroomlight Hit', 'Shroomlight Place', 'Shroomlight Step', 'Soul Sand Break', 'Soul Sand Fall', 'Soul Sand Hit', 'Soul Sand Place', 'Soul Sand Step', 'Soul Soil Break', 'Soul Soil Fall', 'Soul Soil Hit', 'Soul Soil Place', 'Soul Soil Step', 'Chorus Flower Death', 'Chorus Flower Grow', 'Add Cake Candle', 'Candle Ambient', 'Candle Break', 'Candle Extinguish', 'Candle Fall', 'Candle Hit', 'Candle Place', 'Candle Step', 'Chain Break', 'Chain Fall', 'Chain Hit', 'Chain Place', 'Chain Step', 'Copper Block Break', 'Copper Block Fall', 'Copper Block Hit', 'Copper Block Place', 'Copper Block Step', 'Decorated Pot Break', 'Decorated Pot Fall', 'Decorated Pot Hit', 'Decorated Pot Place', 'Decorated Pot Step', 'Decorated Pot Shatter', 'Decorated Pot Insert Item', 'Decorated Pot Insert Fail', 'Glass Break', 'Glass Fall', 'Glass Hit', 'Glass Place', 'Glass Step', 'Waxed Sign Interact Fail', 'Waxed Hanging Sign Interact Fail', 'Wooden Hanging Sign Break', 'Wooden Hanging Sign Fall', 'Wooden Hanging Sign Hit', 'Wooden Hanging Sign Place', 'Wooden Hanging Sign Step', 'Cherry Hanging Sign Break', 'Cherry Hanging Sign Fall', 'Cherry Hanging Sign Hit', 'Cherry Hanging Sign Place', 'Cherry Hanging Sign Step', 'Bamboo Hanging Sign Break', 'Bamboo Hanging Sign Fall', 'Bamboo Hanging Sign Hit', 'Bamboo Hanging Sign Place', 'Bamboo Hanging Sign Step', 'Nether Wood Hanging Sign Break', 'Nether Wood Hanging Sign Fall', 'Nether Wood Hanging Sign Hit', 'Nether Wood Hanging Sign Place', 'Nether Wood Hanging Sign Step', 'Heavy Core Break', 'Heavy Core Fall', 'Heavy Core Hit', 'Heavy Core Place', 'Heavy Core Step', 'Honey Block Break', 'Honey Block Fall', 'Honey Block Hit', 'Honey Block Place', 'Honey Block Slide', 'Honey Block Step', 'Ladder Break', 'Ladder Fall', 'Ladder Hit', 'Ladder Place', 'Ladder Step', 'Lantern Break', 'Lantern Fall', 'Lantern Hit', 'Lantern Place', 'Lantern Step', 'Metal Break', 'Metal Fall', 'Metal Hit', 'Metal Place', 'Metal Step', 'Nether Bricks Break', 'Nether Bricks Fall', 'Nether Bricks Hit', 'Nether Bricks Place', 'Nether Bricks Step', 'Netherite Block Break', 'Netherite Block Fall', 'Netherite Block Hit', 'Netherite Block Place', 'Netherite Block Step', 'Polished Deepslate Break', 'Polished Deepslate Fall', 'Polished Deepslate Hit', 'Polished Deepslate Place', 'Polished Deepslate Step', 'Polished Tuff Break', 'Polished Tuff Fall', 'Polished Tuff Hit', 'Polished Tuff Place', 'Polished Tuff Step', 'Scaffolding Break', 'Scaffolding Fall', 'Scaffolding Hit', 'Scaffolding Place', 'Scaffolding Step', 'Slime Block Break', 'Slime Block Fall', 'Slime Block Hit', 'Slime Block Place', 'Slime Block Step', 'Trial Spawner About to Spawn Item', 'Trial Spawner Ambient', 'Trial Spawner Ambient (Ominous)', 'Trial Spawner Break', 'Trial Spawner Close Shutter', 'Trial Spawner Detect Player', 'Trial Spawner Eject Item', 'Trial Spawner Fall', 'Trial Spawner Hit', 'Trial Spawner Ominous Activate', 'Trial Spawner Open Shutter', 'Trial Spawner Place', 'Trial Spawner Spawn Item', 'Trial Spawner Spawn Item Begin', 'Trial Spawner Spawn Mob', 'Trial Spawner Step', 'Tuff Bricks Break', 'Tuff Bricks Fall', 'Tuff Bricks Hit', 'Tuff Bricks Place', 'Tuff Bricks Step', 'Vault Activate', 'Vault Ambient', 'Vault Break', 'Vault Close Shutter', 'Vault Deactivate', 'Vault Eject Item', 'Vault Fall', 'Vault Hit', 'Vault Insert Item', 'Vault Insert Item (Fail)', 'Vault Open Shutter', 'Vault Place', 'Vault Reject Rewarded Player', 'Vault Step', 'Wool Break', 'Wool Fall', 'Wool Hit', 'Wool Place', 'Wool Step', 'Mud Bricks Break', 'Mud Bricks Fall', 'Mud Bricks Hit', 'Mud Bricks Place', 'Mud Bricks Step', 'Packed Mud Break', 'Packed Mud Fall', 'Packed Mud Hit', 'Packed Mud Place', 'Packed Mud Step', 'Anvil Break', 'Anvil Destroy', 'Anvil Fall', 'Anvil Hit', 'Anvil Land', 'Anvil Place', 'Anvil Step', 'Anvil Use', 'Blast Furnace Fire Crackle', 'Brewing Stand Brew', 'Campfire Crackle', 'Cartography Table Take Result', 'Composter Empty', 'Composter Fill Success', 'Composter Fill', 'Composter Ready', 'Enchantment Table Use', 'Furnace Fire Crackle', 'Grindstone Use', 'Loom Select Pattern', 'Loom Take Result', 'Smithing Table Use', 'Smoker Smoke', 'Stonecutter Select Recipe', 'Stonecutter Take Result', 'Barrel Close', 'Barrel Open', 'Chest Close', 'Chest Locked', 'Chest Open', 'Chiseled Bookshelf Break', 'Chiseled Bookshelf Fall', 'Chiseled Bookshelf Hit', 'Chiseled Bookshelf Place', 'Chiseled Bookshelf Step', 'Chiseled Bookshelf Insert', 'Chiseled Bookshelf Insert (Enchanted)', 'Chiseled Bookshelf Pick Up', 'Chiseled Bookshelf Pick Up (Enchanted)', 'Ender Chest Close', 'Ender Chest Open', 'Shulker Box Close', 'Shulker Box Open', 'Comparator Click', 'Dispenser Dispense', 'Dispenser Fail', 'Dispenser Launch', 'Wooden Fence Gate Close', 'Wooden Fence Gate Open', 'Cherry Fence Gate Close', 'Cherry Fence Gate Open', 'Bamboo Fence Gate Close', 'Bamboo Fence Gate Open', 'Nether Wood Fence Gate Close', 'Nether Wood Fence Gate Open', 'Iron Door Close', 'Iron Door Open', 'Iron Trapdoor Close', 'Iron Trapdoor Open', 'Lever Click', 'Metal Pressure Plate Click Off', 'Metal Pressure Plate Click On', 'Piston Retract', 'Piston Extend', 'Redstone Torch Burnout', 'Sculk Sensor Break', 'Sculk Sensor Click', 'Sculk Sensor Click Stop', 'Sculk Sensor Fall', 'Sculk Sensor Hit', 'Sculk Sensor Place', 'Sculk Sensor Step', 'Stone Button Click Off', 'Stone Button Click On', 'Stone Pressure Plate Click Off', 'Stone Pressure Plate Click On', 'Tripwire Attach', 'Tripwire Click Off', 'Tripwire Click On', 'Tripwire Detach', 'Wooden Button Click Off', 'Wooden Button Click On', 'Wooden Door Close', 'Wooden Door Open', 'Wooden Pressure Plate Click Off', 'Wooden Pressure Plate Click On', 'Wooden Trapdoor Close', 'Wooden Trapdoor Open', 'Cherry Wood Button Click Off', 'Cherry Wood Button Click On', 'Cherry Wood Door Close', 'Cherry Wood Door Open', 'Cherry Wood Pressure Plate Click Off', 'Cherry Wood Pressure Plate Click On', 'Cherry Wood Trapdoor Close', 'Cherry Wood Trapdoor Open', 'Bamboo Button Click Off', 'Bamboo Button Click On', 'Bamboo Door Close', 'Bamboo Door Open', 'Bamboo Pressure Plate Click Off', 'Bamboo Pressure Plate Click On', 'Bamboo Trapdoor Close', 'Bamboo Trapdoor Open', 'Nether Wood Button Click Off', 'Nether Wood Button Click On', 'Nether Wood Door Close', 'Nether Wood Door Open', 'Nether Wood Pressure Plate Click Off', 'Nether Wood Pressure Plate Click On', 'Nether Wood Trapdoor Close', 'Nether Wood Trapdoor Open', 'Bass Drum', 'Banjo', 'Bass', 'Bell', 'Bit', 'Chime', 'Cow Bell', 'Didgeridoo', 'Flute', 'Guitar', 'Harp', 'Pling', 'Hat', 'Snare Drum', 'Iron Xylophone', 'Xylophone', 'Creeper (Note Block)', 'Ender Dragon (Note Block)', 'Piglin (Note Block)', 'Skeleton (Note Block)', 'Wither Skeleton (Note Block)', 'Zombie (Note Block)', 'Beacon Activate', 'Beacon Ambient', 'Beacon Deactivate', 'Beacon Power Select', 'Bell Resonate', 'Bell Use', 'Conduit Activate', 'Conduit Ambient (Short)', 'Conduit Ambient', 'Conduit Attack Target', 'Conduit Deactivate', 'End Gateway Spawn', 'End Portal Fill Frame', 'End Portal Spawn', 'Lodestone Break', 'Lodestone Fall', 'Lodestone Hit', 'Lodestone Lock Compass', 'Lodestone Place', 'Lodestone Step', 'Nether Portal Ambient', 'Nether Portal Travel', 'Nether Portal Trigger', 'Respawn Anchor Ambient', 'Respawn Anchor Charge', 'Respawn Anchor Deplete', 'Respawn Anchor Set Spawnpoint', 'Cave Ambience', 'Lightning Impact', 'Lightning Thunder', 'Raid Horn', 'Underwater Ambience', 'Underwater Ambience Additions', 'Underwater Ambience Additions (Rare)', 'Underwater Ambience Additions (Very Rare)', 'Weather Rain', 'Weather Rain Above', 'Basalt Deltas Ambience', 'Basalt Deltas Ambience Additions', 'Basalt Deltas Ambience (Mood)', 'Crimson Forest Ambience', 'Crimson Forest Ambience Additions', 'Crimson Forest Ambience (Mood)', 'Nether Wastes Ambience', 'Nether Wastes Ambience Additions', 'Nether Wastes Ambience (Mood)', 'Soul Escape', 'Soul Sand Valley Ambience', 'Soul Sand Valley Ambience Additions', 'Soul Sand Valley Ambience (Mood)', 'Warped Forest Ambience', 'Warped Forest Ambience Additions', 'Warped Forest Ambience (Mood)']
POTION_NAME = Literal['Absorption', 'Conduit Power', "Dolphin's Grace", 'Fire Resistance', 'Haste', 'Health Boost', 'Hero of the Village', 'Instant Health', 'Invisibility', 'Jump Boost', 'Luck', 'Night Vision', 'Regeneration', 'Resistance', 'Saturation', 'Slow Falling', 'Strength', 'Speed', 'Water Breathing', 'Bad Omen', 'Blindness', 'Darkness', 'Glowing', 'Hunger', 'Infested', 'Instant Damage', 'Levitation', 'Mining Fatigue', 'Nausea', 'Oozing', 'Poison', 'Raid Omen', 'Slowness', 'Trial Omen', 'Bad Luck', 'Weakness', 'Weaving', 'Wind Charged', 'Wither']
REPEAT_SUBACTION = IF_PLAYER_ACTION | IF_ENTITY_ACTION | IF_GAME_ACTION | IF_VAR_ACTION



"""
A package for making code templates for the DiamondFire Minecraft server.

By Amp
"""

import json
from difflib import get_close_matches
import datetime
from enum import Enum
from amulet_nbt import CompoundTag
from dfpyre.util import *
from dfpyre.items import *
from dfpyre.scriptgen import generate_script, GeneratorFlags
from dfpyre.actiondump import CODEBLOCK_DATA, get_default_tags
from dfpyre.action_literals import *


VARIABLE_TYPES = {'txt', 'comp', 'num', 'item', 'loc', 'var', 'snd', 'part', 'pot', 'g_val', 'vec', 'pn_el', 'bl_tag'}
TEMPLATE_STARTERS = {'event', 'entity_event', 'func', 'process'}
DYNAMIC_CODEBLOCKS = {'func', 'process', 'call_func', 'start_process'}

TARGETS = ['Selection', 'Default', 'Killer', 'Damager', 'Shooter', 'Victim', 'AllPlayers', 'Projectile', 'AllEntities', 'AllMobs', 'LastEntity']
TARGET_CODEBLOCKS = {'player_action', 'entity_action', 'if_player', 'if_entity'}

CODECLIENT_URL = 'ws://localhost:31375'


class Target(Enum):
    SELECTION = 0
    DEFAULT = 1
    KILLER = 2
    DAMAGER = 3
    SHOOTER = 4
    VICTIM = 5
    ALL_PLAYERS = 6
    PROJECTILE = 7
    ALL_ENTITIES = 8
    ALL_MOBS = 9
    LAST_ENTITY = 10

    def get_string_value(self):
        return TARGETS[self.value]

DEFAULT_TARGET = Target.SELECTION


def _convert_args(args):
    return tuple(map(convert_argument, args))


class CodeBlock:
    def __init__(self, codeblock_type: str, action_name: str, args: tuple=(), target: Target=DEFAULT_TARGET, data: dict={}, tags: dict[str, str]={}):
        self.type = codeblock_type
        self.action_name = action_name
        self.args = args
        self.target = target
        self.data = data
        self.tags = tags
    

    @classmethod
    def new_action(cls, codeblock_type: str, action_name: str, args: tuple, tags: dict[str, str], target: Target=DEFAULT_TARGET) -> "CodeBlock":
        args = _convert_args(args)
        return cls(codeblock_type, action_name, args=args, data={'id': 'block', 'block': codeblock_type, 'action': action_name}, tags=tags, target=target)

    @classmethod
    def new_data(cls, codeblock_type: str, data_value: str, args: tuple, tags: dict[str, str]) -> "CodeBlock":
        args = _convert_args(args)
        return cls(codeblock_type, 'dynamic', args=args, data={'id': 'block', 'block': codeblock_type, 'data': data_value}, tags=tags)

    @classmethod
    def new_conditional(cls, codeblock_type: str, action_name: str, args: tuple, tags: dict[str, str], inverted: bool, target: Target=DEFAULT_TARGET) -> "CodeBlock":
        args = _convert_args(args)
        data = {'id': 'block', 'block': codeblock_type, 'action': action_name}
        if inverted:
            data['attribute'] = 'NOT'
        return cls(codeblock_type, action_name, args=args, data=data, tags=tags, target=target)

    @classmethod
    def new_repeat(cls, action_name: str, args: tuple, tags: dict[str, str], sub_action: str|None, inverted: bool) -> "CodeBlock":
        args = _convert_args(args)
        data = {'id': 'block', 'block': 'repeat', 'action': action_name}
        if inverted:
            data['attribute'] = 'NOT'
        if sub_action is not None:
            data['subAction'] = sub_action
        return cls('repeat', action_name, args=args, data=data, tags=tags)

    @classmethod
    def new_else(cls) -> "CodeBlock":
        return cls('else', 'else', data={'id': 'block', 'block': 'else'})

    @classmethod
    def new_bracket(cls, direction: Literal['open', 'close'], bracket_type: Literal['norm', 'repeat']) -> "CodeBlock":
        return cls('bracket', 'bracket', data={'id': 'bracket', 'direct': direction, 'type': bracket_type})


    def __repr__(self) -> str:
        if self.action_name == 'dynamic':
            return f'CodeBlock({self.data["block"]}, {self.data["data"]})'
        if self.action_name == 'else':
            return 'CodeBlock(else)'
        if 'block' in self.data:
            return f'CodeBlock({self.data["block"]}, {self.action_name})'
        return f'CodeBlock(bracket, {self.data["type"]}, {self.data["direct"]})'


    def build(self, include_tags: bool=True) -> dict:
        """
        Builds a properly formatted block from a CodeBlock object.
        """
        built_block = self.data.copy()
        
        # add target if necessary ('Selection' is the default when 'target' is blank)
        if self.type in TARGET_CODEBLOCKS and self.target != DEFAULT_TARGET:
            built_block['target'] = self.target.get_string_value()
        
        # add items into args
        final_args = [arg.format(slot) for slot, arg in enumerate(self.args) if arg.type in VARIABLE_TYPES]
        already_applied_tags: dict[str, dict] = {a['item']['data']['tag']: a for a in final_args if a['item']['id'] == 'bl_tag'}
        
        # check for unrecognized name, add tags
        if self.type not in {'bracket', 'else'}:
            if self.action_name not in CODEBLOCK_DATA[self.type]:
                _warn_unrecognized_name(self.type, self.action_name)
            
            elif include_tags:
                tags = _get_codeblock_tags(self.type, self.action_name, self.tags)
                for i, tag_data in enumerate(tags):
                    already_applied_tag_data = already_applied_tags.get(tag_data['item']['data']['tag'])
                    if already_applied_tag_data is not None:
                        tags[i] = already_applied_tag_data
                
                if len(final_args) + len(tags) > 27:
                    final_args = final_args[:(27-len(tags))]  # trim list if over 27 elements
                final_args.extend(tags)  # add tags to end

        built_block['args'] = {'items': final_args}
        return built_block


def _warn_unrecognized_name(codeblock_type: str, codeblock_name: str):
    close = get_close_matches(codeblock_name, CODEBLOCK_DATA[codeblock_type].keys())
    if close:
        warn(f'Code block name "{codeblock_name}" not recognized. Did you mean "{close[0]}"?')
    else:
        warn(f'Code block name "{codeblock_name}" not recognized. Try spell checking or retyping without spaces.')


def _check_applied_tags(tags: list[dict], applied_tags: dict[str, str], codeblock_name: str) -> dict[str, str]:
    if len(applied_tags) > 0 and len(tags) == 0:
        warn(f'Action "{codeblock_name}" does not have any tags, but still received {len(applied_tags)}.')
        return {}
    
    valid_tags = {}
    tags_formatted = {t['name']: t for t in tags}
    for name, option in applied_tags.items():
        if name not in tags_formatted:
            tag_names_joined = '\n'.join(map(lambda s: '    - '+s, tags_formatted.keys()))
            warn(f'Tag "{name}" does not exist for action "{codeblock_name}". Available tags:\n{tag_names_joined}')
        elif option not in tags_formatted[name]['options']:
            options_joined = '\n'.join(map(lambda s: '    - '+s, tags_formatted[name]['options']))
            warn(f'Tag "{name}" does not have the option "{option}". Available tag options:\n{options_joined}')
        else:
            valid_tags[name] = option
    return valid_tags


def _reformat_codeblock_tags(tags: list[dict], codeblock_type: str, codeblock_action: str, applied_tags: dict[str, str]) -> list[dict]:
    """
    Turns tag objects into DiamondFire formatted tag items.
    """

    def format_tag(option: str, name: str):
        return {
            'item': {
                'id': 'bl_tag',
                'data': {'option': option, 'tag': name, 'action': codeblock_action, 'block': codeblock_type}
            },
            'slot': tag_item['slot']
        }
    
    valid_applied_tags = _check_applied_tags(tags, applied_tags, codeblock_action)
    reformatted_tags = []
    for tag_item in tags:
        tag_name = tag_item['name']
        tag_option = tag_item['default']
        if tag_name in valid_applied_tags:
            tag_option = valid_applied_tags[tag_name]

        new_tag_item = format_tag(tag_option, tag_name)
        reformatted_tags.append(new_tag_item)
    return reformatted_tags


def _get_codeblock_tags(codeblock_type: str, codeblock_name: str, applied_tags: dict[str, str]) -> list[dict]:
    """
    Get tags for the specified codeblock type and name.
    """
    action_data = CODEBLOCK_DATA[codeblock_type][codeblock_name]
    if 'deprecatedNote' in action_data:
        warn(f'Action "{codeblock_name}" is deprecated: {action_data["deprecatedNote"]}')
    tags = action_data['tags']
    return _reformat_codeblock_tags(tags, codeblock_type, codeblock_name, applied_tags)


def _generate_template_item(template_code: str, name: str, author: str) -> Item:
    now = datetime.datetime.now()

    template_item = Item('yellow_shulker_box')
    template_item.set_name(f'&x&f&f&5&c&0&0>> &x&f&f&c&7&0&0{name}')
    template_item.set_lore([
        f'&8Author: {author}',
        f'&8Date: {now.strftime("%Y-%m-%d")}',
        '',
        '&7This template was generated by &6pyre&7.',
        '&7https://github.com/Amp63/pyre'
    ])
    
    custom_data_tag = CompoundTag({
        'PublicBukkitValues': CompoundTag({
            'hypercube:codetemplatedata': StringTag(f'{{"author":"{author}","name":"{name}","version": 1,"code":"{template_code}"}}'),
            'hypercube:pyre_creation_timestamp': DoubleTag(now.timestamp())
        })
    })
    template_item.set_component('minecraft:custom_data', custom_data_tag)

    return template_item


class DFTemplate:
    """
    Represents a DiamondFire code template.
    """
    def __init__(self, codeblocks: list[CodeBlock], author: str='pyre'):
        self.codeblocks = codeblocks
        self.author = author
    

    def _get_template_name(self):
        first_block_data = self.codeblocks[0].data
        if 'data' in first_block_data:
            name = first_block_data['data']
            return name if name else 'Unnamed Template'
        return first_block_data['block'] + '_' + first_block_data['action']


    def __repr__(self) -> str:
        return f'DFTemplate(name: "{self._get_template_name()}", author: "{self.author}", codeblocks: {len(self.codeblocks)})'


    @staticmethod
    def from_code(template_code: str, preserve_item_slots: bool=True):
        """
        Create a template object from an existing template code.

        :param str template_code: The base64 string to create a template from.
        :param bool preserve_item_slots: If True, the positions of items within chests will be saved.
        """
        template_dict = json.loads(df_decode(template_code))
        codeblocks: list[CodeBlock] = []
        for block_dict in template_dict['blocks']:
            block_tags = get_default_tags(block_dict.get('block'), block_dict.get('action'))
            if 'args' in block_dict:
                block_args = []
                for item_dict in block_dict['args']['items']:
                    if item_dict['item'].get('id') == 'bl_tag':
                        tag_data = item_dict['item']['data']
                        block_tags[tag_data['tag']] = tag_data['option']
                    parsed_item = item_from_dict(item_dict, preserve_item_slots)
                    if parsed_item is not None:
                        block_args.append(parsed_item)
            block_target = Target(TARGETS.index(block_dict['target'])) if 'target' in block_dict else DEFAULT_TARGET

            codeblock_type = block_dict.get('block')

            if codeblock_type is None:
                codeblock = CodeBlock.new_bracket(block_dict['direct'], block_dict['type'])
            if codeblock_type == 'else':
                codeblock = CodeBlock.new_else()
            elif codeblock_type in DYNAMIC_CODEBLOCKS:
                codeblock = CodeBlock.new_data(codeblock_type, block_dict['data'], block_args, block_tags)
            elif 'action' in block_dict:
                codeblock = CodeBlock.new_action(codeblock_type, block_dict['action'], block_args, block_tags, block_target)
            codeblocks.append(codeblock)
        
        return DFTemplate(codeblocks)

    
    def insert(self, insert_codeblocks: CodeBlock|list[CodeBlock], index: int=-1) -> "DFTemplate":
        """
        Insert `insert_codeblocks` into this template at `index`.

        :param CodeBlock|list[CodeBlock] insert_codeblocks: The block(s) to insert
        :param int index: The index to insert at.
        :return: self
        """
        if isinstance(insert_codeblocks, list):
            insert_codeblocks = list(flatten(insert_codeblocks))
            if index == -1:
                self.codeblocks.extend(insert_codeblocks)
            else:
                self.codeblocks[index:index+len(insert_codeblocks)] = insert_codeblocks
        elif isinstance(insert_codeblocks, CodeBlock):
            if index == -1:
                index = len(self.codeblocks)
            self.codeblocks.insert(index, insert_codeblocks)
        else:
            raise PyreException('Expected CodeBlock or list[CodeBlock] to insert.')
        return self


    def build(self, include_tags: bool=True) -> str:
        """
        Build this template.

        :param bool include_tags: If True, include item tags in code blocks. Otherwise omit them.
        :return: String containing encoded template data.
        """
        template_dict_blocks = [codeblock.build(include_tags) for codeblock in self.codeblocks]
        template_dict = {'blocks': template_dict_blocks}
        first_block = template_dict_blocks[0]
        if first_block['block'] not in TEMPLATE_STARTERS:
            warn('Template does not start with an event, function, or process.')

        json_string = json.dumps(template_dict, separators=(',', ':'))
        return df_encode(json_string)
    

    def build_and_send(self, include_tags: bool=True) -> int:
        """
        Builds this template and sends it to DiamondFire automatically.
        
        :param bool include_tags: If True, include item tags in code blocks. Otherwise omit them.
        """
        template_code = self.build(include_tags)
        template_item = _generate_template_item(template_code, self._get_template_name(), self.author)
        return template_item.send_to_minecraft()
    
    
    def generate_script(self, output_path: str, indent_size: int=4, literal_shorthand: bool=True, var_shorthand: bool=False, preserve_slots: bool=False):
        """
        Generate an equivalent python script for this template.

        :param str output_path: The file path to write the script to.
        :param int indent_size: The multiple of spaces to add when indenting lines.
        :param bool literal_shorthand: If True, `text` and `num` items will be written as strings and ints respectively.
        :param bool var_shorthand: If True, all variables will be written using variable shorthand.
        :param bool preserve_slots: If True, the positions of items within chests will be saved.
        """
        flags = GeneratorFlags(indent_size, literal_shorthand, var_shorthand, preserve_slots)
        with open(output_path, 'w', encoding='utf-8') as f:
            f.write(generate_script(self, flags))


def _assemble_template(starting_block: CodeBlock, codeblocks: list[CodeBlock], author: str|None) -> DFTemplate:
    """
    Create a DFTemplate object from a starting block and a list of codeblocks.
    `codeblocks` can contain nested lists of CodeBlock objects, so it must be flattened.
    """
    if author is None:
        author = 'pyre'
    template_codeblocks = [starting_block] + list(flatten(codeblocks))  # flatten codeblocks list and insert starting block
    return DFTemplate(template_codeblocks, author)


def player_event(event_name: EVENT_ACTION, codeblocks: list[CodeBlock]=(), author: str|None=None) -> DFTemplate:
    """
    Represents a Player Event codeblock.

    :param str event_name: The name of the event. (Ex: "Join")
    :param list[CodeBlock] codeblocks: The list of codeblocks in this template.
    :param str|None author: The author of this template.
    """
    starting_block = CodeBlock.new_action('event', event_name, (), {})
    return _assemble_template(starting_block, codeblocks, author)


def entity_event(event_name: ENTITY_EVENT_ACTION, codeblocks: list[CodeBlock]=[], author: str|None=None) -> DFTemplate:
    """
    Represents an Entity Event codeblock.

    :param str event_name: The name of the event. (Ex: "EntityDmg")
    :param list[CodeBlock] codeblocks: The list of codeblocks in this template.
    :param str|None author: The author of this template.
    """
    starting_block = CodeBlock.new_action('entity_event', event_name, (), {})
    return _assemble_template(starting_block, codeblocks, author)


def function(function_name: str, *args, tags: dict[str, str]={}, codeblocks: list[CodeBlock]=[], author: str|None=None) -> DFTemplate:
    """
    Represents a Function codeblock.

    :param str event_name: The name of the function.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    :param list[CodeBlock] codeblocks: The list of codeblocks in this template.
    :param str|None author: The author of this template.
    """
    starting_block = CodeBlock.new_data('func', function_name, args, tags)
    return _assemble_template(starting_block, codeblocks, author)


def process(process_name: str, *args, tags: dict[str, str]={}, codeblocks: list[CodeBlock]=[], author: str|None=None) -> DFTemplate:
    """
    Represents a Process codeblock.

    :param str event_name: The name of the process.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    :param list[CodeBlock] codeblocks: The list of codeblocks in this template.
    :param str|None author: The author of this template.
    """
    starting_block = CodeBlock.new_data('process', process_name, args, tags)
    return _assemble_template(starting_block, codeblocks, author)


def call_function(function_name: str, *args) -> CodeBlock:
    """
    Represents a Call Function codeblock.

    :param str event_name: The name of the function.
    :param tuple args: The argument items to include.
    """
    return CodeBlock.new_data('call_func', function_name, args, {})


def start_process(process_name: str, *args, tags: dict[str, str]={}) -> CodeBlock:
    """
    Represents a Call Function codeblock.

    :param str event_name: The name of the function.
    :param tuple args: The argument items to include.
    """
    return CodeBlock.new_data('start_process', process_name, args, tags)


def player_action(action_name: PLAYER_ACTION_ACTION, *args, target: Target=DEFAULT_TARGET, tags: dict[str, str]={}) -> CodeBlock:
    """
    Represents a Player Action codeblock.

    :param str action_name: The name of the action.
    :param tuple args: The argument items to include.
    :param Target target: The target for the action.
    :param dict[str, str] tags: The tags to include.
    """
    return CodeBlock.new_action('player_action', action_name, args, tags, target=target)


def entity_action(action_name: ENTITY_ACTION_ACTION, *args, target: Target=DEFAULT_TARGET, tags: dict[str, str]={}) -> CodeBlock:
    """
    Represents an Entity Action codeblock.

    :param str action_name: The name of the action.
    :param tuple args: The argument items to include.
    :param Target target: The target for the action.
    :param dict[str, str] tags: The tags to include.
    """
    return CodeBlock.new_action('entity_action', action_name, args, tags, target=target)


def game_action(action_name: GAME_ACTION_ACTION, *args, tags: dict[str, str]={}) -> CodeBlock:
    """
    Represents a Game Action codeblock.

    :param str action_name: The name of the action.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    """
    return CodeBlock.new_action('game_action', action_name, args, tags)


def if_player(action_name: IF_PLAYER_ACTION, *args, target: Target=DEFAULT_TARGET, tags: dict[str, str]={}, inverted: bool=False, codeblocks: list[CodeBlock]=[]) -> list[CodeBlock]:
    """
    Represents an If Player codeblock.

    :param str action_name: The name of the condition.
    :param tuple args: The argument items to include.
    :param Target target: The target for the condition.
    :param dict[str, str] tags: The tags to include.
    :param bool inverted: Whether the condition should be inverted.
    :param list[CodeBlock] codeblocks: The list of codeblocks inside the brackets.
    """
    return [
        CodeBlock.new_conditional('if_player', action_name, args, tags, inverted, target),
        CodeBlock.new_bracket('open', 'norm')
    ] + list(codeblocks) + [
        CodeBlock.new_bracket('close', 'norm')
    ]

def if_entity(action_name: IF_ENTITY_ACTION, *args, target: Target=DEFAULT_TARGET, tags: dict[str, str]={}, inverted: bool=False, codeblocks: list[CodeBlock]=[]) -> list[CodeBlock]:
    """
    Represents an If Entity codeblock.

    :param str action_name: The name of the condition.
    :param tuple args: The argument items to include.
    :param Target target: The target for the condition.
    :param dict[str, str] tags: The tags to include.
    :param bool inverted: Whether the condition should be inverted.
    :param list[CodeBlock] codeblocks: The list of codeblocks inside the brackets.
    """
    return [
        CodeBlock.new_conditional('if_entity', action_name, args, tags, inverted, target),
        CodeBlock.new_bracket('open', 'norm')
    ] + list(codeblocks) + [
        CodeBlock.new_bracket('close', 'norm')
    ]


def if_game(action_name: IF_GAME_ACTION, *args, tags: dict[str, str]={}, inverted: bool=False, codeblocks: list[CodeBlock]=[]) -> list[CodeBlock]:
    """
    Represents an If Game codeblock.

    :param str action_name: The name of the condition.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    :param bool inverted: Whether the condition should be inverted.
    :param list[CodeBlock] codeblocks: The list of codeblocks inside the brackets.
    """
    return [
        CodeBlock.new_conditional('if_game', action_name, args, tags, inverted),
        CodeBlock.new_bracket('open', 'norm')
    ] + list(codeblocks) + [
        CodeBlock.new_bracket('close', 'norm')
    ]


def if_variable(action_name: IF_VAR_ACTION, *args, tags: dict[str, str]={}, inverted: bool=False, codeblocks: list[CodeBlock]=[]) -> list[CodeBlock]:
    """
    Represents an If Variable codeblock.

    :param str action_name: The name of the condition.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    :param bool inverted: Whether the condition should be inverted.
    :param list[CodeBlock] codeblocks: The list of codeblocks inside the brackets.
    """
    return [
        CodeBlock.new_conditional('if_var', action_name, args, tags, inverted),
        CodeBlock.new_bracket('open', 'norm')
    ] + list(codeblocks) + [
        CodeBlock.new_bracket('close', 'norm')
    ]


def else_(codeblocks: list[CodeBlock]=[]) -> list[CodeBlock]:
    """
    Represents an Else codeblock.

    :param list[CodeBlock] codeblocks: The list of codeblocks inside the brackets.
    """
    return [
        CodeBlock.new_else(),
        CodeBlock.new_bracket('open', 'norm')
    ] + list(codeblocks) + [
        CodeBlock.new_bracket('close', 'norm')
    ]


def repeat(action_name: REPEAT_ACTION, *args, tags: dict[str, str]={}, sub_action: REPEAT_SUBACTION|None=None, inverted: bool=False, codeblocks: list[CodeBlock]=[]) -> CodeBlock:
    """
    Represents a Repeat codeblock.

    :param str action_name: The name of the action.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    :param str|None sub_action: The sub-action for the repeat action (Only relevant for `While`)
    :param bool inverted: Whether the sub-action condition should be inverted.
    :param list[CodeBlock] codeblocks: The list of codeblocks inside the brackets.
    """
    return [
        CodeBlock.new_repeat(action_name, args, tags, sub_action, inverted),
        CodeBlock.new_bracket('open', 'repeat')
    ] + list(codeblocks) + [
        CodeBlock.new_bracket('close', 'repeat')
    ]


def control(action_name: CONTROL_ACTION, *args, tags: dict[str, str]={}) -> CodeBlock:
    """
    Represents a Control codeblock.

    :param str action_name: The name of the action.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    """
    return CodeBlock.new_action('control', action_name, args, tags)


def select_object(action_name: SELECT_OBJ_ACTION, *args, tags: dict[str, str]={}) -> CodeBlock:
    """
    Represents a Select Object codeblock.

    :param str action_name: The name of the action.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    """
    return CodeBlock.new_action('select_obj', action_name, args, tags) 


def set_variable(action_name: SET_VAR_ACTION, *args, tags: dict[str, str]={}) -> CodeBlock:
    """
    Represents a Set Variable codeblock.

    :param str action_name: The name of the action.
    :param tuple args: The argument items to include.
    :param dict[str, str] tags: The tags to include.
    """
    return CodeBlock.new_action('set_var', action_name, args, tags)



import os
import json
from typing import TypedDict
from dfpyre.util import warn


ACTIONDUMP_PATH = os.path.join(os.path.dirname(__file__), 'data/actiondump_min.json')

CODEBLOCK_NAME_LOOKUP = {
    'PLAYER ACTION': 'player_action',
    'ENTITY ACTION': 'entity_action',
    'GAME ACTION': 'game_action',
    'SET VARIABLE': 'set_var',
    'IF PLAYER': 'if_player',
    'IF ENTITY': 'if_entity',
    'IF GAME': 'if_game',
    'IF VARIABLE': 'if_var',
    'REPEAT': 'repeat',
    'SELECT OBJECT': 'select_obj',
    'CONTROL': 'control',
    'PLAYER EVENT': 'event',
    'ENTITY EVENT': 'entity_event',
    'FUNCTION': 'func',
    'CALL FUNCTION': 'call_func',
    'PROCESS': 'process',
    'START PROCESS': 'start_process',
}


class ActiondumpResult(TypedDict):
    codeblock_data: dict[str, dict]
    game_value_names: list[str]
    sound_names: list[str]
    potion_names: list[str]


def get_action_tags(action_data: dict) -> list[dict]:
    action_tags = []
    for tag_data in action_data['tags']:
        options = [o['name'] for o in tag_data['options']]
        converted_tag_data = {
            'name': tag_data['name'],
            'options': options,
            'default': tag_data['defaultOption'],
            'slot': tag_data['slot']
        }
        action_tags.append(converted_tag_data)
    return action_tags


def parse_actiondump() -> ActiondumpResult:
    codeblock_data = {n: {} for n in CODEBLOCK_NAME_LOOKUP.values()}
    codeblock_data['else'] = {'tags': []}

    if not os.path.exists(ACTIONDUMP_PATH):
        warn('data.json not found -- Item tags and error checking will not work.')
        return {}, set()
    
    with open(ACTIONDUMP_PATH, 'r', encoding='utf-8') as f:
        actiondump = json.loads(f.read())
    for action_data in actiondump['actions']:
        action_tags = get_action_tags(action_data)
        parsed_action_data = {'tags': action_tags, 'required_rank': 'None'}
        if dep_note := action_data['icon']['deprecatedNote']:
            parsed_action_data['deprecatedNote'] = ' '.join(dep_note)
        
        required_rank = action_data['icon']['requiredRank']
        if required_rank:
            parsed_action_data['required_rank'] = required_rank
        
        codeblock_name = CODEBLOCK_NAME_LOOKUP[action_data['codeblockName']]
        codeblock_data[codeblock_name][action_data['name']] = parsed_action_data
        if aliases := action_data['aliases']:
            alias_data = parsed_action_data.copy()
            alias_data['alias'] = action_data['name']
            for alias in aliases:
                codeblock_data[codeblock_name][alias] = alias_data
    
    game_value_names: list[str] = []
    for game_value in actiondump['gameValues']:
        game_value_names.append(game_value['icon']['name'])

    sound_names: list[str] = []
    for sound in actiondump['sounds']:
        sound_names.append(sound['icon']['name'])
    
    potion_names: list[str] = []
    for potion in actiondump['potions']:
        potion_names.append(potion['icon']['name'])
    
    return ActiondumpResult(
        codeblock_data=codeblock_data,
        game_value_names=game_value_names,
        sound_names=sound_names,
        potion_names=potion_names
    )


def get_default_tags(codeblock_type: str|None, codeblock_action: str|None) -> dict[str, str]:
    if codeblock_type is None or codeblock_action is None:
        return {}
    return {t['name']: t['default'] for t in CODEBLOCK_DATA[codeblock_type][codeblock_action]['tags']}


ACTIONDUMP = parse_actiondump()

CODEBLOCK_DATA = ACTIONDUMP['codeblock_data']



"""
Class definitions for code items.
"""

from enum import Enum
import re
from typing import Literal, Any
import websocket
from mcitemlib.itemlib import Item as NbtItem, MCItemlibException
from amulet_nbt import DoubleTag, StringTag, CompoundTag
from dfpyre.style import is_ampersand_coded, ampersand_to_minimessage
from dfpyre.util import PyreException, warn, COL_SUCCESS, COL_ERROR, COL_RESET
from dfpyre.action_literals import GAME_VALUE_NAME, SOUND_NAME, POTION_NAME


NUMBER_REGEX = r'^-?\d*\.?\d+$'
VAR_SHORTHAND_REGEX = r'^\$([gsli]) (.+)$'
VAR_SCOPES = {'g': 'unsaved', 's': 'saved', 'l': 'local', 'i': 'line'}

CODECLIENT_URL = 'ws://localhost:31375'


def convert_argument(arg: Any):
    if type(arg) in {int, float}:
        return Number(arg)
    elif isinstance(arg, str):
        shorthand_match: re.Match = re.match(VAR_SHORTHAND_REGEX, arg)
        if shorthand_match:
            scope = VAR_SCOPES[shorthand_match.group(1)]
            return Variable(shorthand_match.group(2), scope)
        return String(arg)
    return arg


def _add_slot(d: dict, slot: int|None):
    if slot is not None:
        d['slot'] = slot


class String:
    """
    Represents a DiamondFire string object. (`txt`)
    """
    type = 'txt'

    def __init__(self, value: str, slot: int|None=None):
        self.value = value
        self.slot = slot
    
    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type, "data": {"name": self.value}}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict
    
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.value}")'
    
Str = String  # String alias


class Text:
    """
    Represents a DiamondFire styled text object (`comp`)
    """
    type = 'comp'

    def __init__(self, value: str, slot: int|None=None):
        if is_ampersand_coded(value):
            value = ampersand_to_minimessage(value)
        self.value = value
        self.slot = slot
    
    def format(self, slot: int|None):
      formatted_dict = {"item": {"id": self.type, "data": {"name": self.value}}}
      _add_slot(formatted_dict, self.slot or slot)
      return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.value}")'


class Number:
    """
    Represents a DiamondFire number object.
    """
    type = 'num'

    def __init__(self, num: int|float|str, slot: int|None=None):
        self.value = num
        self.slot = slot
    
    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type, "data": {"name": str(self.value)}}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.value})'

Num = Number  # Number alias


class Item(NbtItem):
    """
    Represents a Minecraft item.
    """
    type = 'item'

    def __init__(self, item_id: str, count: int=1, slot: int | None=None):
        super().__init__(item_id, count)
        self.slot = slot

    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type, "data": {"item": self.get_snbt()}}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.get_id()}, {self.get_count()})'
    
    def set_tag(self, tag_name: str, tag_value: str|int|float|String|Number):
        """
        Add a DiamondFire custom tag to this item.
        """
        if isinstance(tag_value, String):
            tag = StringTag(tag_value.value)
        elif isinstance(tag_value, str):
            tag = StringTag(tag_value)
        elif isinstance(tag_value, Number):
            tag = DoubleTag(float(tag_value.value))
        elif isinstance(tag_value, (int, float)):
            tag = DoubleTag(float(tag_value))
        
        try:
            custom_data_tag = self.get_component('minecraft:custom_data')
            if 'PublicBukkitValues' in custom_data_tag:
                pbv_tag = custom_data_tag['PublicBukkitValues']
            else:
                pbv_tag = CompoundTag()
        except MCItemlibException:
            custom_data_tag = CompoundTag()
            pbv_tag = CompoundTag()
        
        custom_data_tag['PublicBukkitValues'] = pbv_tag
        
        pbv_tag[f'hypercube:{tag_name}'] = tag
        self.set_component('minecraft:custom_data', custom_data_tag)
    
    def get_tag(self, tag_name: str) -> str|float|None:
        """
        Get a DiamondFire custom tag from this item.
        """
        try:
            custom_data_tag = self.get_component('minecraft:custom_data')
        except MCItemlibException:
            return None
        
        if 'PublicBukkitValues' not in custom_data_tag:
            return None
        
        pbv_tag = custom_data_tag['PublicBukkitValues']
        df_tag_value = pbv_tag.get(f'hypercube:{tag_name}')
        if df_tag_value is None:
            return None
        
        if isinstance(df_tag_value, DoubleTag):
            return float(df_tag_value)
        if isinstance(df_tag_value, StringTag):
            return str(df_tag_value)
    
    def remove_tag(self, tag_name: str):
        """
        Remove a DiamondFire custom tag from this item.
        """
        custom_data_tag = self.get_component('minecraft:custom_data')
        pbv_tag = custom_data_tag['PublicBukkitValues']
        del pbv_tag[f'hypercube:{tag_name}']

        return True
    
    def send_to_minecraft(self):
        """
        Sends this item to Minecraft automatically.
        """
        try:
            ws = websocket.WebSocket()
            ws.connect(CODECLIENT_URL)
            print(f'{COL_SUCCESS}Connected.{COL_RESET}')

            command = f'give {self.get_snbt()}'
            ws.send(command)
            ws.close()

            print(f'{COL_SUCCESS}Item sent to client successfully.{COL_RESET}')
            return 0
            
        except Exception as e:
            if isinstance(e, ConnectionRefusedError):
                print(f'{COL_ERROR}Could not connect to CodeClient API. Possible problems:')
                print(f'    - Minecraft is not open')
                print(f'    - CodeClient is not installed (get it here: https://modrinth.com/mod/codeclient)')
                print(f'    - CodeClient API is not enabled (enable it in CodeClient general settings){COL_RESET}')
                return 1
            
            print(f'Connection failed: {e}')
            return 2


class Location:
    """
    Represents a DiamondFire location object.
    """
    type = 'loc'

    def __init__(self, x: float=0, y: float=0, z: float=0, pitch: float=0, yaw: float=0, slot: int | None=None):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)
        self.pitch = float(pitch)
        self.yaw = float(yaw)
        self.slot = slot
    
    def format(self, slot: int|None):
        formatted_dict =  {"item": {
            "id": self.type,
            "data": {
                "isBlock": False,
                "loc": {
                    "x": self.x,
                    "y": self.y,
                    "z": self.z,
                    "pitch": self.pitch,
                    "yaw": self.yaw
                }
            }
        }}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.x}, {self.y}, {self.z}, {self.pitch}, {self.yaw})'

Loc = Location  # Location alias


class Variable:
    """
    Represents a DiamondFire variable object.
    """
    type = 'var'

    def __init__(self, name: str, scope: Literal['unsaved', 'game', 'saved', 'local', 'line']='unsaved', slot: int | None=None):
        self.name = name

        if scope == 'game':
            scope = 'unsaved'
        self.scope = scope
        
        self.slot = slot

    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type,"data": {"name": self.name, "scope": self.scope}}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.scope}, "{self.name}")'

Var = Variable  # Variable alias


class Sound:
    """
    Represents a DiamondFire sound object.
    """
    type = 'snd'

    def __init__(self, name: SOUND_NAME, pitch: float=1.0, vol: float=2.0, slot: int | None=None):
        if name not in set(SOUND_NAME.__args__):
            warn(f'Sound name "{name}" not found.')
        
        self.name = name
        self.pitch = pitch
        self.vol = vol
        self.slot = slot

    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type,"data": {"sound": self.name, "pitch": self.pitch, "vol": self.vol}}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(pitch: {self.pitch}, volume: {self.vol})'

Snd = Sound  # Sound alias


class Particle:
    """
    Represents a DiamondFire particle object.
    """
    type = 'part'
    def __init__(self, particle_data: dict, slot: int | None=None):
        self.particle_data = particle_data
        self.slot = slot
    
    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type, "data": self.particle_data}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.particle_data})'


class Potion:
    """
    Represents a DiamondFire potion object.
    """
    type = 'pot'

    def __init__(self, name: POTION_NAME, dur: int=1000000, amp: int=0, slot: int | None=None):
        if name not in set(POTION_NAME.__args__):
            warn(f'Potion name "{name}" not found.')
        
        self.name = name
        self.dur = dur
        self.amp = amp
        self.slot = slot
    
    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type,"data": {"pot": self.name, "dur": self.dur, "amp": self.amp}}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(effect: {self.name}, duration: {self.dur}, amplifier: {self.amp})'

Pot = Potion  # Potion alias


class GameValue:
    """
    Represents a DiamondFire game value object.
    """
    type = 'g_val'

    def __init__(self, name: GAME_VALUE_NAME, target: str='Default', slot: int | None=None):
        if name not in set(GAME_VALUE_NAME.__args__):
            warn(f'Game value name "{name}" not found.')
        
        self.name = name
        self.target = target
        self.slot = slot
    
    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type, "data": {"type": self.name, "target": self.target}}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.name}, target: {self.target})'


class Vector:
    """
    Represents a DiamondFire vector object.
    """
    type = 'vec'

    def __init__(self, x: float=0.0, y: float=0.0, z: float=0.0, slot: int | None=None):
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)
        self.slot = slot
    
    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type, "data": {"x": self.x, "y": self.y, "z": self.z}}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.x}, {self.y}, {self.z})'

Vec = Vector  # Vector alias



PARAMETER_TYPE_LOOKUP = ['txt', 'comp', 'num', 'loc', 'vec', 'snd', 'part', 'pot', 'item', 'any', 'var', 'list', 'dict']

class ParameterType(Enum):
    STRING = 0
    TEXT = 1
    NUMBER = 2
    LOCATION = 3
    VECTOR = 4
    SOUND = 5
    PARTICLE = 6
    POTION_EFFECT = 7
    ITEM = 8
    ANY = 9
    VAR = 10
    LIST = 11
    DICT = 12

    def get_string_value(self) -> str:
        return PARAMETER_TYPE_LOOKUP[self.value]

class Parameter:
    """
    Represents a DiamondFire parameter object.
    """
    type = 'pn_el'

    def __init__(self, name: str, param_type: ParameterType, plural: bool=False, optional: bool=False, 
                 description: str="", note: str="", default_value=None, slot: int | None=None):
        self.name = name
        self.param_type = param_type
        self.plural = plural
        self.optional = optional
        self.description = description
        self.note = note
        self.default_value = convert_argument(default_value)
        self.slot = slot
    
    def format(self, slot: int):
        formatted_dict = {"item": {
            "id": self.type,
            "data": {
                "name": self.name,
                "type": self.param_type.get_string_value(),
                "plural": self.plural,
                "optional": self.optional,
            }}
        }
        _add_slot(formatted_dict, self.slot or slot)

        if self.description:
            formatted_dict['item']['data']['description'] = self.description
        if self.note:
            formatted_dict['item']['data']['note'] = self.note
        if self.default_value is not None:
            if not self.optional:
                warn(f'For parameter "{self.name}": Default value cannot be set if optional is False.')
            elif self.plural:
                warn(f'For parameter "{self.name}": Default value cannot be set while plural is True.')
            else:
                formatted_dict['item']['data']['default_value'] = self.default_value.format(None)['item']
        
        return formatted_dict
    
    def __repr__(self) -> str:
        raw_type = str(self.param_type).partition('.')[2]
        return f'{self.__class__.__name__}({self.name}, type: {raw_type})'


class _Tag:
    """
    Represents a CodeBlock action tag.
    """
    type = 'bl_tag'

    def __init__(self, tag_data: dict, slot: int | None=None):
        self.tag_data = tag_data
        self.slot = slot
    
    def format(self, slot: int|None):
        formatted_dict = {"item": {"id": self.type, "data": self.tag_data}}
        _add_slot(formatted_dict, self.slot or slot)
        return formatted_dict

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.tag_data})'


def item_from_dict(item_dict: dict, preserve_item_slots: bool) -> Any:
    item_id = item_dict['item']['id']
    item_data = item_dict['item']['data']
    item_slot = item_dict['slot'] if preserve_item_slots else None

    if item_id == 'item':
        item = Item.from_snbt(item_data['item'])
        item.slot = item_slot
        return item
    
    elif item_id == 'txt':
        return String(item_data['name'], item_slot)
    
    elif item_id == 'comp':
        return Text(item_data['name'], item_slot)
    
    elif item_id == 'num':
        num_value = item_data['name']
        if re.match(NUMBER_REGEX, num_value):
            num_value = float(item_data['name'])
            if num_value % 1 == 0:
                num_value = int(num_value)
            return Number(num_value, item_slot)
        return Number(num_value, item_slot)
    
    elif item_id == 'loc':
        item_loc = item_data['loc']
        return Location(item_loc['x'], item_loc['y'], item_loc['z'], item_loc['pitch'], item_loc['yaw'], item_slot)
    
    elif item_id == 'var':
        return Variable(item_data['name'], item_data['scope'], item_slot)
    
    elif item_id == 'snd':
        return Sound(item_data['sound'], item_data['pitch'], item_data['vol'], item_slot)
    
    elif item_id == 'part':
        return Particle(item_data, item_slot)
    
    elif item_id == 'pot':
        return Potion(item_data['pot'], item_data['dur'], item_data['amp'], item_slot)
    
    elif item_id == 'g_val':
        return GameValue(item_data['type'], item_data['target'], item_slot)
    
    elif item_id == 'vec':
        return Vector(item_data['x'], item_data['y'], item_data['z'], item_slot)
    
    elif item_id == 'pn_el':
        description = item_data.get('description') or ''
        note = item_data.get('note') or ''
        param_type = ParameterType(PARAMETER_TYPE_LOOKUP.index(item_data['type']))
        if item_data['optional']:
            if 'default_value' in item_data:
                default_value_dict = {'item': item_data['default_value'], 'slot': None}
                default_value_item = item_from_dict(default_value_dict, preserve_item_slots)
                return Parameter(item_data['name'], param_type, item_data['plural'], True, description, note, default_value_item, item_slot)
            return Parameter(item_data['name'], param_type, item_data['plural'], True, description, note, slot=item_slot)
        return Parameter(item_data['name'], param_type, item_data['plural'], False, description, note, slot=item_slot)
    
    elif item_id == 'bl_tag':
        if 'variable' in item_data:
            return _Tag(item_data, item_slot)
    
    elif item_id == 'hint':  # Ignore hints
        return
    
    else:
        raise PyreException(f'Unrecognized item id `{item_id}`')



import base64
import gzip


COL_WARN = '\x1b[33m'
COL_RESET = '\x1b[0m'
COL_SUCCESS = '\x1b[32m'
COL_ERROR = '\x1b[31m'


class PyreException(Exception):
    pass


def warn(message: str):
    print(f'{COL_WARN}! WARNING ! {message}{COL_RESET}')


def df_encode(json_string: str) -> str:
    """
    Encodes a stringified json.
    """
    encoded_string = gzip.compress(json_string.encode('utf-8'))
    return base64.b64encode(encoded_string).decode('utf-8')


def df_decode(encoded_string: str) -> str:
    return gzip.decompress(base64.b64decode(encoded_string.encode('utf-8'))).decode('utf-8')


def flatten(nested_list: list):
    """
    Flattens a list.
    """
    for item in nested_list:
        if isinstance(item, list):
            yield from flatten(item)
        else:
            yield item



import dataclasses
import re
from dfpyre.items import *
from dfpyre.actiondump import get_default_tags


SCRIPT_START = '''from dfpyre import *\n\n'''

TEMPLATE_FUNCTION_LOOKUP = {
    'event': 'player_event',
    'entity_event': 'entity_event',
    'func': 'function',
    'process': 'process',
    'call_func': 'call_function',
    'start_process': 'start_process',
    'player_action': 'player_action',
    'game_action': 'game_action',
    'entity_action': 'entity_action',
    'if_player': 'if_player',
    'if_var': 'if_variable',
    'if_game': 'if_game',
    'if_entity': 'if_entity',
    'else': 'else_',
    'repeat': 'repeat',
    'control': 'control',
    'select_obj': 'select_object',
    'set_var': 'set_variable'
}

TARGET_CODEBLOCKS = {'player_action', 'entity_action', 'if_player', 'if_entity'}
CONTAINER_CODEBLOCKS = {'event', 'entity_event', 'func', 'process', 'if_player', 'if_entity', 'if_game', 'if_var', 'else', 'repeat'}
VAR_SCOPES = {'unsaved': 'g', 'saved': 's', 'local': 'l', 'line': 'i'}


@dataclasses.dataclass
class GeneratorFlags:
    indent_size: int
    literal_shorthand: bool
    var_shorthand: bool
    preserve_slots: bool


def item_to_string(class_name: str, i: Item, slot_argument: str):
    i.nbt.pop('DF_NBT', None)
    stripped_id = i.get_id().replace('minecraft:', '')
    if set(i.nbt.keys()) == {'id', 'count'}:
        if i.get_count() == 1:
            return f'{class_name}("{stripped_id}"{slot_argument})'
        return f'{class_name}("{stripped_id}", {i.get_count()}{slot_argument})'
    return f'{class_name}.from_snbt("""{i.get_snbt()}""")'


def argument_item_to_string(flags: GeneratorFlags, arg_item: object) -> str:
    class_name = arg_item.__class__.__name__
    has_slot = arg_item.slot is not None and flags.preserve_slots
    slot_argument = f', slot={arg_item.slot}' if has_slot else ''

    if isinstance(arg_item, Item):
        return item_to_string(class_name, arg_item, slot_argument)
    
    if isinstance(arg_item, String):
        value = arg_item.value.replace('\n', '\\n')
        if not has_slot and flags.literal_shorthand:
            return f'"{value}"'
        return f'{class_name}("{value}"{slot_argument})'
    
    if isinstance(arg_item, Text):
        value = arg_item.value.replace('\n', '\\n')
        return f'{class_name}("{value}"{slot_argument})'
    
    if isinstance(arg_item, Number):
        if not re.match(NUMBER_REGEX, str(arg_item.value)):
            return f'{class_name}("{arg_item.value}"{slot_argument})' 
        if not has_slot and flags.literal_shorthand:
            return str(arg_item.value)
        return f'{class_name}({arg_item.value}{slot_argument})'
    
    if isinstance(arg_item, Location):
        loc_components = [arg_item.x, arg_item.y, arg_item.z]
        if arg_item.pitch != 0:
            loc_components.append(arg_item.pitch)
        if arg_item.yaw != 0:
            loc_components.append(arg_item.yaw)
        return f'{class_name}({", ".join(str(c) for c in loc_components)}{slot_argument})'
    
    if isinstance(arg_item, Variable):
        if not has_slot and flags.var_shorthand:
            return f'"${VAR_SCOPES[arg_item.scope]} {arg_item.name}"'
        if arg_item.scope == 'unsaved':
            return f'{class_name}("{arg_item.name}"{slot_argument})'
        return f'{class_name}("{arg_item.name}", "{arg_item.scope}"{slot_argument})'
    
    if isinstance(arg_item, Sound):
        return f'{class_name}("{arg_item.name}", {arg_item.pitch}, {arg_item.vol}{slot_argument})'
    
    if isinstance(arg_item, Particle):
        return f'{class_name}({arg_item.particle_data})'
    
    if isinstance(arg_item, Potion):
        return f'{class_name}("{arg_item.name}", {arg_item.dur}, {arg_item.amp}{slot_argument})'
    
    if isinstance(arg_item, GameValue):
        if arg_item.target == 'Default':
            return f'{class_name}("{arg_item.name}"{slot_argument})'
        return f'{class_name}("{arg_item.name}", "{arg_item.target}"{slot_argument})'
    
    if isinstance(arg_item, Parameter):
        param_type_class_name = arg_item.param_type.__class__.__name__
        param_args = [f'"{arg_item.name}"', f'{param_type_class_name}.{arg_item.param_type.name}']
        if arg_item.plural:
            param_args.append('plural=True')
        if arg_item.optional:
            param_args.append('optional=True')
            if arg_item.default_value is not None:
                param_args.append(f'default_value={argument_item_to_string(flags, arg_item.default_value)}')
        if arg_item.description:
            param_args.append(f'description="{arg_item.description}"')
        if arg_item.note:
            param_args.append(f'note="{arg_item.note}"')
        return f'{class_name}({", ".join(param_args)}{slot_argument})'
    
    if isinstance(arg_item, Vector):
        return f'{class_name}({arg_item.x}, {arg_item.y}, {arg_item.z}{slot_argument})'


def add_script_line(flags: GeneratorFlags, script_lines: list[str], indent_level: int, line: str, add_comma: bool=True):
    added_line = ' '*flags.indent_size*indent_level + line
    if add_comma and indent_level > 0:
        added_line += ','
    script_lines.append(added_line)


def generate_script(template, flags: GeneratorFlags) -> str:
    indent_level = 0
    script_lines = []

    def remove_comma_from_last_line():
        script_lines[-1] = script_lines[-1][:-1]

    for codeblock in template.codeblocks:
        # Handle closing brackets
        if codeblock.type == 'bracket':
            if codeblock.data['direct'] == 'close':
                remove_comma_from_last_line()
                indent_level -= 1
                add_script_line(flags, script_lines, indent_level, '])')
            continue

        # Get codeblock function and start its arguments with the action
        function_name = TEMPLATE_FUNCTION_LOOKUP[codeblock.type]
        function_args = [f'"{codeblock.action_name}"']

        # Set function or process name if necessary
        if codeblock.action_name == 'dynamic':
            function_args[0] = f'"{codeblock.data["data"]}"'
        
        # Convert argument objects to valid Python strings
        codeblock_args = [argument_item_to_string(flags, i) for i in codeblock.args]
        if codeblock_args:
            function_args.extend(codeblock_args)
        
        # Add target if necessary
        if function_name in TARGET_CODEBLOCKS and codeblock.target.name != 'SELECTION':
            function_args.append(f'target=Target.{codeblock.target.name}')
        
        # Add tags
        if codeblock.tags:
            default_tags = get_default_tags(codeblock.data.get('block'), codeblock.action_name)
            written_tags = {t: o for t, o in codeblock.tags.items() if default_tags[t] != o}
            if written_tags:
                function_args.append(f'tags={str(written_tags)}')
        
        # Add sub-action for repeat
        if codeblock.data.get('subAction'):
            function_args.append(f'sub_action="{codeblock.data["subAction"]}"')
        
        # Add inversion for NOT
        if codeblock.data.get('attribute') == 'NOT':
            function_args.append('inverted=True')

        if codeblock.type in CONTAINER_CODEBLOCKS:
            if codeblock.type == 'else':
                line = f'{function_name}(['
            elif codeblock.type in {'event', 'entity_event'}:
                line = f'{function_name}({", ".join(function_args)}, ['  # omit `codeblocks=` when we don't need it
            else:
                line = f'{function_name}({", ".join(function_args)}, codeblocks=['
            add_script_line(flags, script_lines, indent_level, line, False)
            indent_level += 1
        else:
            line = f'{function_name}({", ".join(function_args)})'
            add_script_line(flags, script_lines, indent_level, line)
    
    remove_comma_from_last_line()
    indent_level -= 1
    add_script_line(flags, script_lines, indent_level, '])')  # add final closing brackets
    return SCRIPT_START + '\n'.join(script_lines)



import re
from mcitemlib.style import STYLE_CODE_REGEX, FORMAT_CODES, StyledString

def is_ampersand_coded(s: str) -> bool:
    return bool(re.match(STYLE_CODE_REGEX, s))


def ampersand_to_minimessage(ampersand_code: str) -> str:
    ampersand_code = ampersand_code.replace('&r', '<reset>')  # bad but should work most of the time
    styled_string = StyledString.from_codes(ampersand_code)
    formatted_string_list = []
    for substring in styled_string.substrings:
        formatted_substring_list = []
        for style_type, value in substring.data.items():
            if style_type in FORMAT_CODES.values() and value:
                formatted_substring_list.append(f'<{style_type}>')
            if style_type == 'color':
                formatted_substring_list.append(f'<{value}>')
        
        formatted_substring_list.append(substring.data['text'])
        formatted_string_list.append(''.join(formatted_substring_list))
    return ''.join(formatted_string_list)